# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from typing import Callable, Optional

import equinox as eqx
import jax
import matplotlib.pyplot as plt
import numpy as np
import optax
import tqdm
from jax import numpy as jnp, random as jrd
from jaxtyping import Array
from loguru import logger

from ._utils import make_model, mse_loss_fn, predict, read_data, save_model


def _train_test_data_split(
    X: Array, Y: Array, batch_size: int, test_size: float = 0.2
) -> tuple[Array, Array, Array, Array]:
    """Split the data into training and testing sets.

    Parameters
    ----------
    X : Array
        Input data
    Y : Array
        Output data
    batch_size : int
        batch size for training
    test_size : float, optional
        fraction of the data to use for testing, by default 0.2

    Returns
    -------
    tuple[Array, Array, Array, Array]
        training and testing sets
    """
    n = len(X)
    indices = np.random.permutation(n)
    split = int(n * (1 - test_size))
    split = (split // batch_size) * batch_size
    train_indices = indices[:split]
    test_indices = indices[split:]

    train_X = X[train_indices]
    train_Y = Y[train_indices]
    test_X = X[test_indices]
    test_Y = Y[test_indices]

    return train_X, test_X, train_Y, test_Y


def train_regressor(
    *,
    input_keys: list[str],
    output_keys: list[str],
    width_size: int,
    depth: int,
    batch_size: int,
    data_path: str,
    checkpoint_path: Optional[str] = None,
    epochs: int = 50,
    validation_split: float = 0.2,
    learning_rate: float = 1e-3,
    train_in_log: bool = False,
) -> None:
    """Train the model to approximate the log of the VT function.

    input_keys : list[str]
        list of input keys
    output_keys : list[str]
        list of output keys
    width_size : int
        width size of the model
    depth : int
        depth of the model
    batch_size : int
        batch size for training
    data_path : str
        path to the data
    checkpoint_path : Optional[str]
        path to save the model, by default None
    epochs : int
        number of epochs to train the model
    validation_split : float
        fraction of the data to use for validation, by default 0.2
    learning_rate : float
        learning rate for the optimizer, by default 1e-3
    train_in_log : bool
        whether to train the model in log space, by default False

    Raises
    ------
    ValueError
        if checkpoint path does not end with :code:`.hdf5`
    """
    if checkpoint_path is None:
        raise ValueError("No checkpoint path provided, model will not be saved.")
    if not checkpoint_path.endswith(".hdf5"):
        raise ValueError("Checkpoint path must end with .hdf5")

    optimizer = optax.adam(learning_rate=learning_rate)

    @eqx.filter_jit
    def make_step(
        model: eqx.nn.MLP,
        x: Array,
        y: Array,
        opt_state: optax.OptState,
    ) -> tuple[eqx.nn.MLP, optax.OptState, Array]:
        """Make a step in the optimization process.

        Parameters
        ----------
        model : eqx.nn.MLP
            Model to approximate the log of the VT function
        x : Array
            input data
        y : Array
            output data
        opt_state : optax.OptState
            optimizer state

        Returns
        -------
        tuple[eqx.nn.MLP, optax.OptState, Array]
            optimizer state
        """
        loss, grads = mse_loss_fn(model, x, y)
        updates, opt_state = optimizer.update(grads, opt_state)
        model = eqx.apply_updates(model, updates)
        return model, opt_state, loss

    def train_step(
        model: eqx.nn.MLP,
        x: Array,
        y: Array,
        opt_state: optax.OptState,
    ) -> tuple[eqx.nn.MLP, optax.OptState, Array]:
        """Train the model for one epoch.

        Parameters
        ----------
        model : eqx.nn.MLP
            Model to approximate the log of the VT function
        x : Array
            input data
        y : Array
            output data
        opt_state : optax.OptState
            optimizer state

        Returns
        -------
        tuple[eqx.nn.MLP, optax.OptState, Array]
            optimizer state
        """
        model, opt_state, loss = make_step(model, x, y, opt_state)
        return model, opt_state, loss

    df = read_data(data_path, keys=input_keys + output_keys)

    data_X = jax.device_put(df[input_keys].to_numpy(), may_alias=True)
    data_Y = jax.device_put(df[output_keys].to_numpy(), may_alias=True)

    if train_in_log:
        data_Y = jnp.log(data_Y)
        data_Y = jnp.where(
            jnp.isneginf(data_Y), -jnp.finfo(jnp.result_type(float)).tiny, data_Y
        )

    train_X, test_X, train_Y, test_Y = _train_test_data_split(
        data_X,
        data_Y,
        batch_size,
        test_size=validation_split,
    )

    logger.info("Input Keys: " + ", ".join(input_keys))
    logger.info("Output Keys: " + ", ".join(output_keys))
    logger.info("Width Size: " + str(width_size))
    logger.info("Depth: " + str(depth))
    logger.info("Data Path: " + data_path)
    logger.info("Checkpoint Path: " + checkpoint_path)
    logger.info("Train Size: " + str(len(train_X)))
    logger.info("Test Size: " + str(len(test_X)))
    logger.info("Validation Split: " + str(validation_split))
    logger.info("Batch Size: " + str(batch_size))
    logger.info("Epochs: " + str(epochs))
    logger.info("Learning Rate: " + str(learning_rate))

    model = make_model(
        key=jrd.PRNGKey(np.random.randint(0, 2**32 - 1)),
        input_layer=len(input_keys),
        output_layer=len(output_keys),
        width_size=width_size,
        depth=depth,
    )

    model: Callable = eqx.filter_checkpoint(model)  # type: ignore

    opt_state = optimizer.init(eqx.filter(model, eqx.is_inexact_array))

    loss_vals = []
    val_loss_vals = []

    with tqdm.tqdm(range(epochs), unit="epochs") as pbar:
        for epoch in pbar:
            epoch_loss = jnp.zeros(())
            for i in range(0, len(train_X), batch_size):
                x = train_X[i : i + batch_size]
                y = train_Y[i : i + batch_size]

                model, opt_state, loss = train_step(model, x, y, opt_state)
                epoch_loss += loss
                pbar.set_postfix({"epoch": epoch + 1, "loss": f"{loss:.5E}"})

            loss = epoch_loss / (len(train_X) // batch_size)
            loss_vals.append(loss)

            val_loss, _ = mse_loss_fn(model, test_X, test_Y)
            val_loss_vals.append(val_loss)
            pbar.set_description(
                f"Epoch {epoch + 1}/{epochs}, Loss: {loss:.5E}, Val Loss: {val_loss:.5E}"
            )
            if epoch % 10 == 0 or epoch == epochs - 1:
                logger.info(
                    f"Epoch {epoch + 1}/{epochs}, Loss: {loss:.5E}, Val Loss: {val_loss:.5E}"
                )

    if checkpoint_path is not None:
        save_model(
            filepath=checkpoint_path,
            datafilepath=data_path,
            model=model,
            names=input_keys,
            is_log=train_in_log,
        )  # type: ignore
        plt.plot(loss_vals, label="loss")
        plt.plot(val_loss_vals, label="val loss")
        plt.yscale("log")
        plt.xlabel("Epoch")
        plt.ylabel("Average loss per epoch")
        plt.legend()
        plt.tight_layout()
        plt.savefig(checkpoint_path + "_loss.png")
        plt.close("all")

        Y_hat = predict(model, data_X)
        total_loss = jnp.square(data_Y - Y_hat)
        total_loss = np.asarray(total_loss)  # type: ignore
        quantiles = np.quantile(total_loss, [0.05, 0.5, 0.95])
        total_loss = total_loss.tolist()  # type: ignore
        total_loss = sorted(total_loss)  # type: ignore

        plt.plot(total_loss, label="loss per data instance")
        plt.axhline(quantiles[0], color="red", linestyle="--", label="5% quantile")
        plt.axhline(quantiles[1], color="green", linestyle="--", label="50% quantile")
        plt.axhline(quantiles[2], color="blue", linestyle="--", label="95% quantile")
        plt.yscale("log")
        plt.xlabel("data instance")
        plt.ylabel("Loss per data instance")
        plt.legend()
        plt.tight_layout()
        plt.savefig(checkpoint_path + "_total_loss.png")
        plt.close("all")



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


#
"""Provides implementation of various transformations using
:class:`~numpyro.distributions.transforms.Transform`.
"""

from jax import numpy as jnp
from jaxtyping import Array
from numpyro.distributions import constraints
from numpyro.distributions.transforms import (
    AbsTransform,
    AffineTransform,
    biject_to,
    ComposeTransform,
    OrderedTransform,
    PowerTransform,
    SigmoidTransform,
    Transform,
)

from ..utils.transformations import (
    chirp_mass,
    delta_m,
    delta_m_to_symmetric_mass_ratio,
    m1_q_to_m2,
    m2_q_to_m1,
    mass_ratio,
    Mc_eta_to_m1_m2,
    symmetric_mass_ratio,
    total_mass,
)
from .constraints import (
    decreasing_vector,
    increasing_vector,
    mass_ratio_mass_sandwich,
    mass_sandwich,
    positive_decreasing_vector,
    positive_increasing_vector,
    strictly_decreasing_vector,
    strictly_increasing_vector,
)


__all__ = [
    "ComponentMassesAndRedshiftToDetectedMassAndRedshift",
    "ComponentMassesToChirpMassAndDelta",
    "ComponentMassesToChirpMassAndSymmetricMassRatio",
    "ComponentMassesToMassRatioAndSecondaryMass",
    "ComponentMassesToPrimaryMassAndMassRatio",
    "ComponentMassesToTotalMassAndMassRatio",
    "DeltaToSymmetricMassRatio",
    "PrimaryMassAndMassRatioToComponentMassesTransform",
    "SourceMassAndRedshiftToDetectedMassAndRedshift",
]


class PrimaryMassAndMassRatioToComponentMassesTransform(Transform):
    r"""Transforms a primary mass and mass ratio to component masses.

    .. math::
        f: (m_1, q)\to (m_1, m_1q)

    .. math::
        f^{-1}: (m_1, m_2)\to (m_1, m_2/m_1)
    """

    domain = constraints.independent(
        constraints.interval(
            jnp.zeros((2,)), jnp.array([jnp.finfo(jnp.result_type(float)).max, 1.0])
        ),
        1,
    )
    r""":math:`\mathcal{D}(f) = \mathbb{R}^2_+\times[0, 1]`"""
    codomain = positive_decreasing_vector
    r""":math:`\mathcal{C}(f)=\{(m_1, m_2)\in\mathbb{R}^2_+\mid m_1\geq m_2>0\}`"""

    def __call__(self, x: Array):
        m1, q = jnp.unstack(x, axis=-1)
        m2 = jnp.multiply(m1, q)
        m1m2 = jnp.stack((m1, m2), axis=-1)
        return m1m2

    def _inverse(self, y: Array):
        m1, m2 = jnp.unstack(y, axis=-1)
        q = mass_ratio(m2=m2, m1=m1)
        m1q = jnp.stack((m1, q), axis=-1)
        return m1q

    def log_abs_det_jacobian(self, x: Array, y: Array, intermediates=None):
        r"""
        .. math::
            \ln\left(|\mathrm{det}(J_f)|\right) = \ln(|m_1|)
        """
        m1 = x[..., 0]
        return jnp.log(jnp.abs(m1))

    def tree_flatten(self):
        return (), ((), dict())

    def __eq__(self, other):
        if not isinstance(other, PrimaryMassAndMassRatioToComponentMassesTransform):
            return False
        return self.domain == other.domain


class ComponentMassesToChirpMassAndSymmetricMassRatio(Transform):
    r"""Transforms component masses to chirp mass and symmetric mass ratio.

    .. math::
        f: (m_1, m_2)\to \left(\frac{(m_1m_2)^{3/5}}{(m_1+m_2)^{1/5}}, \frac{m_1m_2}{(m_1+m_2)^{2}}\right)

    .. seealso::

        - :class:`ComponentMassesAndRedshiftToDetectedMassAndRedshift`
        - :class:`ComponentMassesToChirpMassAndDelta`
        - :class:`ComponentMassesToMassRatioAndSecondaryMass`
        - :class:`ComponentMassesToPrimaryMassAndMassRatio`
        - :class:`ComponentMassesToTotalMassAndMassRatio`
    """

    domain = positive_decreasing_vector
    r""":math:`\mathcal{D}(f)=\{(m_1,m_2)\in\mathbb{R}^2_+\mid m_1\geq m_2>0\}`"""
    codomain = constraints.independent(
        constraints.interval(
            jnp.zeros((2,)), jnp.array([jnp.finfo(jnp.result_type(float)).max, 0.25])
        ),
        1,
    )
    r""":math:`\mathcal{C}(f) = \mathbb{R}^2_+\times[0, 0.25]`"""

    def __call__(self, x):
        m1, m2 = jnp.unstack(x, axis=-1)
        Mc = chirp_mass(m1=m1, m2=m2)
        eta = symmetric_mass_ratio(m1=m1, m2=m2)
        return jnp.stack((Mc, eta), axis=-1)

    def _inverse(self, y):
        Mc, eta = jnp.unstack(y, axis=-1)
        m1, m2 = Mc_eta_to_m1_m2(Mc=Mc, eta=eta)
        return jnp.stack((m1, m2), axis=-1)

    def log_abs_det_jacobian(self, x, y, intermediates=None):
        r"""
        .. math::
            \ln\left(|\mathrm{det}(J_f)|\right)=\ln(M_c)+\ln(\eta)+\ln(m_1-m_2)-\ln(m_1+m_2)-\ln(m_1)-\ln(m_2)"""
        m1, m2 = jnp.unstack(x, axis=-1)
        Mc, eta = jnp.unstack(y, axis=-1)
        # The factor of 2 is omitted to align with empirical results from test cases.
        # Further investigation may be required to reconcile theory with implementation.
        log_detJ = jnp.log(Mc) + jnp.log(eta) + jnp.log(m1 - m2)
        log_detJ -= jnp.log(m1 + m2) + jnp.log(m1) + jnp.log(m2)
        return log_detJ

    def tree_flatten(self):
        return (), ((), dict())

    def __eq__(self, other):
        if not isinstance(other, ComponentMassesToChirpMassAndSymmetricMassRatio):
            return False
        return self.domain == other.domain


class DeltaToSymmetricMassRatio(Transform):
    r"""Transforms delta to symmetric mass ratio.

    .. math::
        \eta = f(\delta) = \frac{1-\delta^2}{4}

    .. math::
        \delta = f^{-1}(\eta) = \sqrt{1-4\eta}
    """

    domain = constraints.unit_interval
    r""":math:`\mathcal{D}(f) = [0, 1]`"""
    codomain = constraints.interval(0.0, 0.25)
    r""":math:`\mathcal{C}(f) = [0, 0.25]`"""

    def __call__(self, x):
        delta_sq = jnp.square(x)
        return jnp.multiply(jnp.subtract(1.0, delta_sq), 0.25)

    def _inverse(self, y):
        eta4 = jnp.multiply(4, y)
        return jnp.sqrt(jnp.subtract(1, eta4))

    def log_abs_det_jacobian(self, x, y, intermediates=None):
        r"""
        .. math::
            \ln\left(|\mathrm{det}(J_f)|\right) = \ln(\delta) - \ln(2)
        """
        return jnp.subtract(jnp.log(x), jnp.log(2.0))

    def tree_flatten(self):
        return (), ((), dict())

    def __eq__(self, other):
        return isinstance(other, type(self))


class ComponentMassesToChirpMassAndDelta(Transform):
    r"""Transforms component masses to chirp mass and delta.

    .. math::
        f: (m_1, m_2) \to (M_c, \delta)

    .. seealso::

        - :class:`ComponentMassesAndRedshiftToDetectedMassAndRedshift`
        - :class:`ComponentMassesToChirpMassAndSymmetricMassRatio`
        - :class:`ComponentMassesToMassRatioAndSecondaryMass`
        - :class:`ComponentMassesToPrimaryMassAndMassRatio`
        - :class:`ComponentMassesToTotalMassAndMassRatio`
    """

    domain = positive_decreasing_vector
    r""":math:`\mathcal{D}(f)=\{(m_1,m_2)\in\mathbb{R}^2_+\mid m_1\geq m_2>0\}`"""
    codomain = constraints.independent(
        constraints.interval(
            jnp.zeros(2), jnp.array([jnp.finfo(jnp.result_type(float)).max, 1.0])
        ),
        1,
    )
    r""":math:`\mathcal{C}(f) = \mathbb{R}^2_+\times[0, 1]`"""

    def __call__(self, x):
        m1 = x[..., 0]
        m2 = x[..., 1]
        Mc = chirp_mass(m1=m1, m2=m2)
        delta = delta_m(m1=m1, m2=m2)
        return jnp.stack((Mc, delta), axis=-1)

    def _inverse(self, y):
        Mc, delta = jnp.unstack(y, axis=-1)
        eta = delta_m_to_symmetric_mass_ratio(delta_m=delta)
        m1, m2 = Mc_eta_to_m1_m2(Mc=Mc, eta=eta)
        return jnp.stack((m1, m2), axis=-1)

    def log_abs_det_jacobian(self, x, y, intermediates=None):
        r"""
        .. math::
            \ln\left(|\mathrm{det}(J_f)|\right) = \ln(2M_c) - 2\ln(m_1+m_2)
        """
        m1, m2 = jnp.unstack(x, axis=-1)
        M = total_mass(m1=m1, m2=m2)
        log_Mc = jnp.log(y[..., 0])
        log_detJ = jnp.log(2.0)
        log_detJ = jnp.add(log_detJ, log_Mc)
        log_detJ = jnp.add(log_detJ, jnp.multiply(-2.0, jnp.log(M)))
        return log_detJ

    def tree_flatten(self):
        return (), ((), dict())

    def __eq__(self, other):
        return isinstance(other, type(self))


class SourceMassAndRedshiftToDetectedMassAndRedshift(Transform):
    r"""Transforms source mass and redshift to detected mass and redshift.

    .. math::
        f: (m_{\text{source}}, z) \to (m_{\text{detected}}, z)
    """

    domain = constraints.independent(constraints.positive, 1)
    r""":math:`\mathcal{D}(f) = \mathbb{R}^2_+`"""
    codomain = constraints.independent(constraints.positive, 1)
    r""":math:`\mathcal{C}(f) = \mathbb{R}^2_+`"""

    def __call__(self, x):
        m_source, z = jnp.unstack(x, axis=-1)
        m_detected = m_source * (1 + z)
        return jnp.stack((m_detected, z), axis=-1)

    def _inverse(self, y):
        m_detected, z = jnp.unstack(y, axis=-1)
        m_source = m_detected / (1 + z)
        return jnp.stack((m_source, z), axis=-1)

    def log_abs_det_jacobian(self, x, y, intermediates=None):
        r"""
        .. math::
            \ln\left(|\mathrm{det}(J_f)|\right) = \ln(1+z)
        """
        z = x[..., 1]
        return jnp.log1p(z)

    def tree_flatten(self):
        return (), ((), dict())

    def __eq__(self, other):
        return isinstance(other, type(self))


class ComponentMassesAndRedshiftToDetectedMassAndRedshift(Transform):
    r"""Transforms component masses and redshift to detected masses and redshift.

    .. math::
        f: (m_1, m_2, z) \to (m_{1, \text{detected}}, m_{2, \text{detected}}, z)

    .. seealso::

        - :class:`ComponentMassesToChirpMassAndDelta`
        - :class:`ComponentMassesToChirpMassAndSymmetricMassRatio`
        - :class:`ComponentMassesToMassRatioAndSecondaryMass`
        - :class:`ComponentMassesToPrimaryMassAndMassRatio`
        - :class:`ComponentMassesToTotalMassAndMassRatio`
    """

    domain = constraints.independent(constraints.positive, 1)
    r""":math:`\mathcal{D}(f) = \mathbb{R}^3_+`"""
    codomain = constraints.independent(constraints.positive, 1)
    r""":math:`\mathcal{C}(f) = \mathbb{R}^3_+`"""

    def __call__(self, x):
        m1_source, m2_source, z = jnp.unstack(x, axis=-1)
        m1_detected = m1_source * (1 + z)
        m2_detected = m2_source * (1 + z)
        return jnp.stack((m1_detected, m2_detected, z), axis=-1)

    def _inverse(self, y):
        m1_detected, m2_detected, z = jnp.unstack(y, axis=-1)
        m1_source = m1_detected / (1 + z)
        m2_source = m2_detected / (1 + z)
        return jnp.stack((m1_source, m2_source, z), axis=-1)

    def log_abs_det_jacobian(self, x, y, intermediates=None):
        r"""
        .. math::
            \ln\left(|\mathrm{det}(J_f)|\right) = 2\ln(1+z)
        """
        z = x[..., 2]
        return 2 * jnp.log1p(z)

    def tree_flatten(self):
        return (), ((), dict())

    def __eq__(self, other):
        return isinstance(other, type(self))


class ComponentMassesToPrimaryMassAndMassRatio(Transform):
    r"""Transforms component masses and redshift to primary mass and mass ratio.

    .. math::
        f: (m_1, m_2) \to (m_1, q)

    .. math::
        f^{-1}: (m_1, q) \to (m_1, m_2)

    .. seealso::

        - :class:`ComponentMassesAndRedshiftToDetectedMassAndRedshift`
        - :class:`ComponentMassesToChirpMassAndDelta`
        - :class:`ComponentMassesToChirpMassAndSymmetricMassRatio`
        - :class:`ComponentMassesToMassRatioAndSecondaryMass`
        - :class:`ComponentMassesToTotalMassAndMassRatio`
    """

    domain = positive_decreasing_vector
    r""":math:`\mathcal{D}(f)=\{(m_1,m_2)\in\mathbb{R}^2_+\mid m_1\geq m_2>0\}`"""
    codomain = constraints.independent(
        constraints.open_interval(
            jnp.zeros(2), jnp.array([jnp.finfo(jnp.result_type(float)).max, 1.0])
        ),
        1,
    )
    r""":math:`\mathcal{C}(f) = \mathbb{R}^2_+\times(0, 1)`"""

    def __call__(self, x):
        m1, m2 = jnp.unstack(x, axis=-1)
        q = mass_ratio(m1=m1, m2=m2)
        return jnp.stack((m1, q), axis=-1)

    def _inverse(self, y):
        m1, q = jnp.unstack(y, axis=-1)
        m2 = m1_q_to_m2(m1=m1, q=q)
        return jnp.stack((m1, m2), axis=-1)

    def log_abs_det_jacobian(self, x, y, intermediates=None):
        m1 = x[..., 0]
        return -jnp.log(jnp.abs(m1))

    def tree_flatten(self):
        return (), ((), dict())

    def __eq__(self, other):
        return isinstance(other, type(self))


class ComponentMassesToMassRatioAndSecondaryMass(Transform):
    r"""Transforms component masses and redshift to mass ratio and secondary mass.

    .. math::
        f: (m_1, m_2) \to (q, m_2)

    .. math::
        f^{-1}: (q, m_2) \to (m_1, m_2)

    .. seealso::

        - :class:`ComponentMassesAndRedshiftToDetectedMassAndRedshift`
        - :class:`ComponentMassesToChirpMassAndDelta`
        - :class:`ComponentMassesToChirpMassAndSymmetricMassRatio`
        - :class:`ComponentMassesToPrimaryMassAndMassRatio`
        - :class:`ComponentMassesToTotalMassAndMassRatio`
    """

    domain = positive_decreasing_vector
    r""":math:`\mathcal{D}(f)=\{(m_1,m_2)\in\mathbb{R}^2_+\mid m_1\geq m_2>0\}`"""
    codomain = constraints.independent(
        constraints.interval(
            jnp.zeros(2), jnp.array([1.0, jnp.finfo(jnp.result_type(float)).max])
        ),
        1,
    )
    r""":math:`\mathcal{C}(f) = [0, 1]\times\mathbb{R}_+`"""

    def __call__(self, x):
        m1, m2 = jnp.unstack(x, axis=-1)
        q = mass_ratio(m1=m1, m2=m2)
        return jnp.stack((q, m2), axis=-1)

    def _inverse(self, y):
        q, m2 = jnp.unstack(y, axis=-1)
        m1 = m2_q_to_m1(m2=m2, q=q)
        return jnp.stack((m1, m2), axis=-1)

    def log_abs_det_jacobian(self, x, y, intermediates=None):
        r"""
        .. math::
            \ln\left(|\mathrm{det}(J_f)|\right) = \ln(q) - \ln(m_1)
        """
        m1 = x[..., 0]
        q = y[..., 0]
        return jnp.log(q) - jnp.log(m1)

    def tree_flatten(self):
        return (), ((), dict())

    def __eq__(self, other):
        return isinstance(other, type(self))


class ComponentMassesToTotalMassAndMassRatio(Transform):
    r"""Transforms component masses to total mass and mass ratio.

    .. math::
        f: (m_1, m_2) \to (M, q)

    .. seealso::

        - :class:`ComponentMassesAndRedshiftToDetectedMassAndRedshift`
        - :class:`ComponentMassesToChirpMassAndDelta`
        - :class:`ComponentMassesToChirpMassAndSymmetricMassRatio`
        - :class:`ComponentMassesToMassRatioAndSecondaryMass`
        - :class:`ComponentMassesToPrimaryMassAndMassRatio`
    """

    domain = positive_decreasing_vector
    codomain = constraints.independent(
        constraints.interval(
            jnp.zeros(2), jnp.array([jnp.finfo(jnp.result_type(float)).max, 1.0])
        ),
        1,
    )

    def __call__(self, x):
        m1, m2 = jnp.unstack(x, axis=-1)
        M = total_mass(m1=m1, m2=m2)
        q = mass_ratio(m1=m1, m2=m2)
        return jnp.stack((M, q), axis=-1)

    def _inverse(self, y):
        M, q = jnp.unstack(y, axis=-1)
        safe_q = jnp.where(q == -1.0, 1.0, q)
        m1 = jnp.where(q == -1.0, jnp.inf, M / (1 + safe_q))
        m2 = m1_q_to_m2(m1=m1, q=q)
        return jnp.stack((m1, m2), axis=-1)

    def log_abs_det_jacobian(self, x, y, intermediates=None):
        m1 = x[..., 0]
        q = y[..., 1]
        return jnp.log(1 + q) - jnp.log(m1)

    def tree_flatten(self):
        return (), ((), dict())

    def __eq__(self, other):
        return isinstance(other, type(self))


@biject_to.register(type(positive_decreasing_vector))
@biject_to.register(type(decreasing_vector))
@biject_to.register(type(strictly_decreasing_vector))
def _transform_to_positive_ordered_vector(constraint):
    """Ensure things to be positive and in decreasing order."""
    return ComposeTransform([AbsTransform(), OrderedTransform(), PowerTransform(-1.0)])


@biject_to.register(type(positive_increasing_vector))
@biject_to.register(type(increasing_vector))
@biject_to.register(type(strictly_increasing_vector))
def _transform_to_positive_ordered_vector(constraint):
    """Ensure things to be positive and in decreasing order."""
    return ComposeTransform([AbsTransform(), OrderedTransform()])


@biject_to.register(mass_sandwich)
def _transform_to_mass_sandwich(constraint):
    return ComposeTransform(
        [
            AbsTransform(),
            OrderedTransform(),
            PowerTransform(-1.0),
            SigmoidTransform(),
            AffineTransform(
                loc=constraint.mmin, scale=constraint.mmax - constraint.mmin
            ),
        ]
    )


@biject_to.register(mass_ratio_mass_sandwich)
def _transform_to_mass_sandwich(constraint):
    return ComposeTransform(
        [
            AbsTransform(),
            OrderedTransform(),
            PowerTransform(-1.0),
            SigmoidTransform(),
            AffineTransform(
                loc=jnp.array([constraint.mmin, 0.0]),
                scale=jnp.array([constraint.mmax - constraint.mmin, 1.0]),
            ),
        ]
    )



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0

# source: https://github.com/pyro-ppl/numpyro/blob/3cde93d0f25490b9b90c1c423816c6cfd9ea23ed/test/test_distributions.py

import inspect
import types

import equinox as eqx
import jax
import jax.numpy as jnp
import jax.random as jrd
import numpy as np
import numpyro.distributions as dist
import pytest
import scipy.stats as osp
from jax import lax, vmap
from numpy.testing import assert_allclose
from numpyro.distributions import constraints
from numpyro.distributions.batch_util import vmap_over
from numpyro.distributions.gof import auto_goodness_of_fit, InvalidTest
from numpyro.distributions.transforms import biject_to
from numpyro.distributions.util import multinomial, signed_stick_breaking_tril
from scipy.sparse import csr_matrix

from gwkokab.models import (
    BetaFromMeanVar,
    NPowerlawMGaussian,
    PowerlawPrimaryMassRatio,
    PowerlawRedshift,
    Wysocki2019MassModel,
)
from gwkokab.models.constraints import (
    decreasing_vector,
    increasing_vector,
    mass_ratio_mass_sandwich,
    mass_sandwich,
    positive_decreasing_vector,
    positive_increasing_vector,
    strictly_decreasing_vector,
    strictly_increasing_vector,
)
from gwkokab.models.utils import ScaledMixture


TEST_FAILURE_RATE = 2e-5  # For all goodness-of-fit tests.


def my_kron(A, B):
    D = A[..., :, None, :, None] * B[..., None, :, None, :]
    ds = D.shape
    newshape = (*ds[:-4], ds[-4] * ds[-3], ds[-2] * ds[-1])
    return D.reshape(newshape)


generic_nspmsg = {
    ## rates
    "log_rate_0": -2.0,
    "log_rate_1": -1.0,
    "log_rate_2": 1.5,
    "log_rate_3": 1.0,
    ## powerlaw 0
    "alpha_pl_0": 2.0,
    "beta_pl_0": 3.0,
    "mmin_pl_0": 50.0,
    "mmax_pl_0": 70.0,
    "delta_pl_0": 5,
    ## powerlaw 1
    "alpha_pl_1": -1.5,
    "beta_pl_1": -1.0,
    "mmin_pl_1": 20.0,
    "mmax_pl_1": 100.0,
    "delta_pl_1": 20,
    ## gaussian 0
    "loc_g_0": 70.0,
    "scale_g_0": 2.1,
    "beta_g_0": 3.2,
    "mmin_g_0": 50.0,
    "mmax_g_0": 180.0,
    "delta_g_0": 5,
    ## gaussian 1
    "loc_g_1": 80.0,
    "scale_g_1": 1.1,
    "beta_g_1": 2.2,
    "mmin_g_1": 20.0,
    "mmax_g_1": 180.0,
    "delta_g_1": 5,
    # "use_spin": True,
    "chi1_mean_g": 0.5,
    "chi1_mean_pl": 0.7,
    "chi2_mean_g": 0.2,
    "chi2_mean_pl": 0.6,
    "chi1_variance_g": 0.1,
    "chi1_variance_pl": 0.2,
    "chi2_variance_g": 0.14,
    "chi2_variance_pl": 0.1,
    # "use_tilt": True,
    "cos_tilt_zeta_g_0": 0.5,
    "cos_tilt_zeta_g_1": 0.5,
    "cos_tilt_zeta_pl_0": 0.5,
    "cos_tilt_zeta_pl_1": 0.5,
    "cos_tilt1_scale_g_0": 0.1,
    "cos_tilt1_scale_g_1": 0.3,
    "cos_tilt1_scale_pl_0": 0.1,
    "cos_tilt1_scale_pl_1": 0.3,
    "cos_tilt2_scale_g_0": 0.1,
    "cos_tilt2_scale_g_1": 0.3,
    "cos_tilt2_scale_pl_0": 0.1,
    "cos_tilt2_scale_pl_1": 0.3,
    # "use_eccentricity": True,
    "ecc_loc_pl_0": 0.1,
    "ecc_scale_pl_0": 0.2,
    "ecc_loc_pl_1": 0.1,
    "ecc_scale_pl_1": 0.4,
    "ecc_loc_g_0": 0.1,
    "ecc_scale_g_0": 0.7,
    "ecc_loc_g_1": 0.1,
    "ecc_scale_g_1": 0.6,
    "ecc_low_pl_0": 0.0,
    "ecc_low_g_0": 0.0,
    "ecc_low_pl_1": 0.0,
    "ecc_low_g_1": 0.0,
    "ecc_high_pl_0": 1.0,
    "ecc_high_g_0": 1.0,
    "ecc_high_pl_1": 1.0,
    "ecc_high_g_1": 1.0,
}


generic_npmg = {
    ## rates
    "log_rate_0": -2.0,
    "log_rate_1": -1.0,
    "log_rate_2": 1.5,
    "log_rate_3": 1.0,
    ## powerlaw 0
    "alpha_pl_0": 2.0,
    "beta_pl_0": 3.0,
    "mmin_pl_0": 50.0,
    "mmax_pl_0": 70.0,
    "delta_pl_0": 5,
    ## powerlaw 1
    "alpha_pl_1": -1.5,
    "beta_pl_1": -1.0,
    "mmin_pl_1": 20.0,
    "mmax_pl_1": 100.0,
    "delta_pl_1": 20,
    ## gaussian 0
    "m1_loc_g_0": 70.0,
    "m2_loc_g_0": 30.0,
    "m1_scale_g_0": 2.1,
    "m2_scale_g_0": 3.2,
    "m1_low_g_0": 10.0,
    "m1_high_g_0": 180.0,
    "m2_low_g_0": 10.0,
    "m2_high_g_0": 180.0,
    ## gaussian 1
    "m1_loc_g_1": 80.0,
    "m2_loc_g_1": 20.0,
    "m1_scale_g_1": 1.1,
    "m2_scale_g_1": 2.2,
    "m1_low_g_1": 10.0,
    "m1_high_g_1": 180.0,
    "m2_low_g_1": 10.0,
    "m2_high_g_1": 180.0,
    # "use_spin": True,
    "chi1_mean_g": 0.5,
    "chi1_mean_pl": 0.7,
    "chi2_mean_g": 0.2,
    "chi2_mean_pl": 0.6,
    "chi1_variance_g": 0.1,
    "chi1_variance_pl": 0.2,
    "chi2_variance_g": 0.14,
    "chi2_variance_pl": 0.1,
    # "use_tilt": True,
    "cos_tilt_zeta_g_0": 0.5,
    "cos_tilt_zeta_g_1": 0.5,
    "cos_tilt_zeta_pl_0": 0.5,
    "cos_tilt_zeta_pl_1": 0.5,
    "cos_tilt1_scale_g_0": 0.1,
    "cos_tilt1_scale_g_1": 0.3,
    "cos_tilt1_scale_pl_0": 0.1,
    "cos_tilt1_scale_pl_1": 0.3,
    "cos_tilt2_scale_g_0": 0.1,
    "cos_tilt2_scale_g_1": 0.3,
    "cos_tilt2_scale_pl_0": 0.1,
    "cos_tilt2_scale_pl_1": 0.3,
    # "use_eccentricity": True,
    "ecc_loc_pl_0": 0.1,
    "ecc_scale_pl_0": 0.2,
    "ecc_loc_pl_1": 0.1,
    "ecc_scale_pl_1": 0.4,
    "ecc_loc_g_0": 0.1,
    "ecc_scale_g_0": 0.7,
    "ecc_loc_g_1": 0.1,
    "ecc_scale_g_1": 0.6,
    "ecc_low_pl_0": 0.0,
    "ecc_low_g_0": 0.0,
    "ecc_low_pl_1": 0.0,
    "ecc_low_g_1": 0.0,
    "ecc_high_pl_0": 1.0,
    "ecc_high_g_0": 1.0,
    "ecc_high_pl_1": 1.0,
    "ecc_high_g_1": 1.0,
}


CONTINUOUS = [
    (
        PowerlawPrimaryMassRatio,
        {"alpha": -1.0, "beta": 1.0, "mmin": 10.0, "mmax": 50.0},
    ),
    (
        PowerlawPrimaryMassRatio,
        {"alpha": -0.1, "beta": -8.0, "mmin": 70.0, "mmax": 100.0},
    ),
    (
        PowerlawPrimaryMassRatio,
        {"alpha": -1.4, "beta": 9.0, "mmin": 5.0, "mmax": 100.0},
    ),
    (PowerlawPrimaryMassRatio, {"alpha": 2.0, "beta": 3.0, "mmin": 50.0, "mmax": 70.0}),
    (Wysocki2019MassModel, {"alpha_m": -1.0, "mmin": 10.0, "mmax": 50.0}),
    (Wysocki2019MassModel, {"alpha_m": -2.3, "mmin": 5.0, "mmax": 100.0}),
    (Wysocki2019MassModel, {"alpha_m": 0.7, "mmin": 50.0, "mmax": 100.0}),
    (Wysocki2019MassModel, {"alpha_m": 3.1, "mmin": 70.0, "mmax": 100.0}),
    ######### NPowerlawMGaussian (m1, m2) #########
    (NPowerlawMGaussian, {"N_pl": 1, "N_g": 0, **generic_npmg}),
    (NPowerlawMGaussian, {"N_pl": 0, "N_g": 1, **generic_npmg}),
    (NPowerlawMGaussian, {"N_pl": 1, "N_g": 1, **generic_npmg}),
    (NPowerlawMGaussian, {"N_pl": 1, "N_g": 2, **generic_npmg}),
    (NPowerlawMGaussian, {"N_pl": 2, "N_g": 2, **generic_npmg}),
    ######### NPowerlawMGaussian (m1, m2, chi1, chi2) #########
    (NPowerlawMGaussian, {"N_pl": 1, "N_g": 0, "use_spin": True, **generic_npmg}),
    (NPowerlawMGaussian, {"N_pl": 0, "N_g": 1, "use_spin": True, **generic_npmg}),
    (NPowerlawMGaussian, {"N_pl": 1, "N_g": 1, "use_spin": True, **generic_npmg}),
    (NPowerlawMGaussian, {"N_pl": 1, "N_g": 2, "use_spin": True, **generic_npmg}),
    (NPowerlawMGaussian, {"N_pl": 2, "N_g": 2, "use_spin": True, **generic_npmg}),
    ######### NPowerlawMGaussian (m1, m2, cos_tilt_1, cos_tilt_2) #########
    (NPowerlawMGaussian, {"N_pl": 1, "N_g": 0, "use_tilt": True, **generic_npmg}),
    (NPowerlawMGaussian, {"N_pl": 0, "N_g": 1, "use_tilt": True, **generic_npmg}),
    (NPowerlawMGaussian, {"N_pl": 1, "N_g": 1, "use_tilt": True, **generic_npmg}),
    (NPowerlawMGaussian, {"N_pl": 1, "N_g": 2, "use_tilt": True, **generic_npmg}),
    (NPowerlawMGaussian, {"N_pl": 2, "N_g": 2, "use_tilt": True, **generic_npmg}),
    ######### NPowerlawMGaussian (m1, m2, ecc) #########
    (
        NPowerlawMGaussian,
        {"N_pl": 1, "N_g": 0, "use_eccentricity": True, **generic_npmg},
    ),
    (
        NPowerlawMGaussian,
        {"N_pl": 0, "N_g": 1, "use_eccentricity": True, **generic_npmg},
    ),
    (
        NPowerlawMGaussian,
        {"N_pl": 1, "N_g": 1, "use_eccentricity": True, **generic_npmg},
    ),
    (
        NPowerlawMGaussian,
        {"N_pl": 1, "N_g": 2, "use_eccentricity": True, **generic_npmg},
    ),
    (
        NPowerlawMGaussian,
        {"N_pl": 2, "N_g": 2, "use_eccentricity": True, **generic_npmg},
    ),
    ######### NPowerlawMGaussian (m1, m2, chi1, chi2, cos_tilt_1, cos_tilt_2, ecc) #########
    (
        NPowerlawMGaussian,
        {
            "N_pl": 1,
            "N_g": 0,
            "use_spin": True,
            "use_tilt": True,
            "use_eccentricity": True,
            **generic_npmg,
        },
    ),
    (
        NPowerlawMGaussian,
        {
            "N_pl": 0,
            "N_g": 1,
            "use_spin": True,
            "use_tilt": True,
            "use_eccentricity": True,
            **generic_npmg,
        },
    ),
    (
        NPowerlawMGaussian,
        {
            "N_pl": 1,
            "N_g": 1,
            "use_spin": True,
            "use_tilt": True,
            "use_eccentricity": True,
            **generic_npmg,
        },
    ),
    (
        NPowerlawMGaussian,
        {
            "N_pl": 1,
            "N_g": 2,
            "use_spin": True,
            "use_tilt": True,
            "use_eccentricity": True,
            **generic_npmg,
        },
    ),
    (
        NPowerlawMGaussian,
        {
            "N_pl": 2,
            "N_g": 2,
            "use_spin": True,
            "use_tilt": True,
            "use_eccentricity": True,
            **generic_npmg,
        },
    ),
    (PowerlawRedshift, {"kappa": 0.0, "z_max": 1.0}),
    (PowerlawRedshift, {"kappa": 1.0, "z_max": 2.3}),
    (PowerlawRedshift, {"kappa": 2.7, "z_max": 1.0}),
    (PowerlawRedshift, {"kappa": 0.0, "z_max": 2.3}),
    (BetaFromMeanVar, {"mean": 0.4, "variance": 0.02}),
    (BetaFromMeanVar, {"mean": 0.5, "variance": 0.05}),
]


def gen_values_within_bounds(constraint, size, key=jrd.PRNGKey(11)):
    eps = 1e-6

    if constraint is constraints.boolean:
        return jrd.bernoulli(key, shape=size)
    elif isinstance(constraint, constraints.greater_than):
        return jnp.exp(jrd.normal(key, size)) + constraint.lower_bound + eps
    elif isinstance(constraint, constraints.less_than):
        return constraint.upper_bound - jnp.exp(jrd.normal(key, size)) - eps
    elif constraint == constraints.positive:
        return jnp.exp(jrd.normal(key, size))
    elif isinstance(constraint, constraints.integer_interval):
        lower_bound = jnp.broadcast_to(constraint.lower_bound, size)
        upper_bound = jnp.broadcast_to(constraint.upper_bound, size)
        return jrd.randint(key, size, lower_bound, upper_bound + 1)
    elif isinstance(constraint, constraints.integer_greater_than):
        return constraint.lower_bound + jrd.poisson(key, np.array(5), shape=size)
    elif isinstance(constraint, constraints.interval):
        lower_bound = jnp.broadcast_to(constraint.lower_bound, size)
        upper_bound = jnp.broadcast_to(constraint.upper_bound, size)
        return jrd.uniform(key, size, minval=lower_bound, maxval=upper_bound)
    elif constraint in (constraints.real, constraints.real_vector):
        return jrd.normal(key, size)
    elif constraint is constraints.simplex:
        return osp.dirichlet.rvs(alpha=jnp.ones((size[-1],)), size=size[:-1])
    elif isinstance(constraint, constraints.multinomial):
        n = size[-1]
        return multinomial(
            key, p=jnp.ones((n,)) / n, n=constraint.upper_bound, shape=size[:-1]
        )
    elif constraint is constraints.corr_cholesky:
        return signed_stick_breaking_tril(
            jrd.uniform(
                key, size[:-2] + (size[-1] * (size[-1] - 1) // 2,), minval=-1, maxval=1
            )
        )
    elif constraint is constraints.corr_matrix:
        cholesky = signed_stick_breaking_tril(
            jrd.uniform(
                key, size[:-2] + (size[-1] * (size[-1] - 1) // 2,), minval=-1, maxval=1
            )
        )
        return jnp.matmul(cholesky, jnp.swapaxes(cholesky, -2, -1))
    elif constraint is constraints.lower_cholesky:
        return jnp.tril(jrd.uniform(key, size))
    elif constraint is constraints.positive_definite:
        x = jrd.normal(key, size)
        return jnp.matmul(x, jnp.swapaxes(x, -2, -1))
    elif isinstance(
        constraint,
        (
            constraints.ordered_vector,
            increasing_vector,
            positive_increasing_vector,
            strictly_increasing_vector,
        ),
    ):
        x = jnp.cumsum(jrd.exponential(key, size), -1)
        return x - jrd.normal(key, size[:-1] + (1,))
    elif isinstance(
        constraint,
        (
            decreasing_vector,
            positive_decreasing_vector,
            strictly_decreasing_vector,
        ),
    ):
        x = jnp.cumsum(jrd.exponential(key, size), -1)
        return x[..., ::-1]
    elif isinstance(constraint, constraints.independent):
        return gen_values_within_bounds(constraint.base_constraint, size, key)
    elif constraint is constraints.sphere:
        x = jrd.normal(key, size)
        return x / jnp.linalg.norm(x, axis=-1)
    elif constraint is constraints.l1_ball:
        key1, key2 = jrd.split(key)
        sign = jrd.bernoulli(key1)
        bounds = [0, (-1) ** sign * 0.5]
        return jrd.uniform(key, size, float, *sorted(bounds))
    elif isinstance(constraint, constraints.zero_sum):
        x = jrd.normal(key, size)
        zero_sum_axes = tuple(i for i in range(-constraint.event_dim, 0))
        for axis in zero_sum_axes:
            x -= x.mean(axis)
        return x
    elif isinstance(constraint, mass_sandwich):
        x = jrd.uniform(
            key, size + (2,), minval=constraint.mmin, maxval=constraint.mmax
        )
        x = jnp.sort(x, axis=-1)
        return x
    elif isinstance(constraint, mass_ratio_mass_sandwich):
        x = jrd.normal(key, size)
        x = jnp.abs(x)
        x = jnp.cumsum(x, -1)
        x = x - jrd.normal(key, size[:-1] + (1,))
        x = jnp.reciprocal(x)
        x = jax.nn.sigmoid(x)
        x = x * jnp.array([constraint.mmax - constraint.mmin, 1.0]) + jnp.array(
            [constraint.mmin, 0.0]
        )
        return x
    else:
        raise NotImplementedError("{} not implemented.".format(constraint))


def gen_values_outside_bounds(constraint, size, key=jrd.PRNGKey(11)):
    if constraint is constraints.boolean:
        return jrd.bernoulli(key, shape=size) - 2
    elif isinstance(constraint, constraints.greater_than):
        return constraint.lower_bound - jnp.exp(jrd.normal(key, size))
    elif isinstance(constraint, constraints.less_than):
        return constraint.upper_bound + jnp.exp(jrd.normal(key, size))
    elif constraint == constraints.positive:
        return -jnp.exp(jrd.normal(key, size))
    elif isinstance(constraint, constraints.integer_interval):
        lower_bound = jnp.broadcast_to(constraint.lower_bound, size)
        return jrd.randint(key, size, lower_bound - 1, lower_bound)
    elif isinstance(constraint, constraints.integer_greater_than):
        return constraint.lower_bound - jrd.poisson(key, np.array(5), shape=size)
    elif isinstance(constraint, constraints.interval):
        upper_bound = jnp.broadcast_to(constraint.upper_bound, size)
        return jrd.uniform(key, size, minval=upper_bound, maxval=upper_bound + 1.0)
    elif constraint in [constraints.real, constraints.real_vector]:
        return lax.full(size, np.nan)
    elif constraint is constraints.simplex:
        return osp.dirichlet.rvs(alpha=jnp.ones((size[-1],)), size=size[:-1]) + 1e-2
    elif isinstance(constraint, constraints.multinomial):
        n = size[-1]
        return (
            multinomial(
                key, p=jnp.ones((n,)) / n, n=constraint.upper_bound, shape=size[:-1]
            )
            + 1
        )
    elif constraint is constraints.corr_cholesky:
        return (
            signed_stick_breaking_tril(
                jrd.uniform(
                    key,
                    size[:-2] + (size[-1] * (size[-1] - 1) // 2,),
                    minval=-1,
                    maxval=1,
                )
            )
            + 1e-2
        )
    elif constraint is constraints.corr_matrix:
        cholesky = 1e-2 + signed_stick_breaking_tril(
            jrd.uniform(
                key, size[:-2] + (size[-1] * (size[-1] - 1) // 2,), minval=-1, maxval=1
            )
        )
        return jnp.matmul(cholesky, jnp.swapaxes(cholesky, -2, -1))
    elif constraint is constraints.lower_cholesky:
        return jrd.uniform(key, size)
    elif constraint is constraints.positive_definite:
        return jrd.normal(key, size)
    elif (
        constraint is constraints.ordered_vector
        or constraint is increasing_vector
        or constraint is positive_increasing_vector
    ):
        x = jnp.cumsum(jrd.exponential(key, size), -1)
        return x[..., ::-1]
    elif isinstance(constraint, constraints.independent):
        return gen_values_outside_bounds(constraint.base_constraint, size, key)
    elif constraint is constraints.sphere:
        x = jrd.normal(key, size)
        x = x / jnp.linalg.norm(x, axis=-1, keepdims=True)
        return 2 * x
    elif constraint is constraints.l1_ball:
        key1, key2 = jrd.split(key)
        sign = jrd.bernoulli(key1)
        bounds = [(-1) ** sign * 1.1, (-1) ** sign * 2]
        return jrd.uniform(key, size, float, *sorted(bounds))
    elif isinstance(constraint, constraints.zero_sum):
        x = jrd.normal(key, size)
        return x
    elif isinstance(constraint, (mass_ratio_mass_sandwich, mass_sandwich)):
        x = jrd.normal(key, size + (2,))
        x = -jnp.abs(x)
        return x
    elif isinstance(
        constraint, (strictly_increasing_vector, strictly_decreasing_vector)
    ):
        x = jnp.full(size, -1.0)
        return x
    elif isinstance(constraint, (decreasing_vector, positive_decreasing_vector)):
        x = jnp.cumsum(jrd.exponential(key, size), -1)
        return x
    else:
        raise NotImplementedError("{} not implemented.".format(constraint))


@pytest.mark.parametrize("jax_dist_cls, params", CONTINUOUS)
@pytest.mark.parametrize("prepend_shape", [(), (2,), (2, 3)])
def test_dist_shape(jax_dist_cls, params, prepend_shape):
    if jax_dist_cls.__name__ in ("PowerlawPeak",):
        pytest.skip(reason=f"{jax_dist_cls.__name__} does not provide sample method")
    if isinstance(jax_dist_cls, types.FunctionType):
        if jax_dist_cls.__name__ in ("NSmoothedPowerlawMSmoothedGaussian",):
            pytest.skip(
                reason=f"{jax_dist_cls.__name__} does not provide sample method"
            )
    jax_dist = jax_dist_cls(**params)
    rng_key = jrd.PRNGKey(0)
    expected_shape = prepend_shape + jax_dist.batch_shape + jax_dist.event_shape
    samples = jax_dist.sample(key=rng_key, sample_shape=prepend_shape)
    assert jnp.shape(samples) == expected_shape


@pytest.mark.parametrize("jax_dist, params", CONTINUOUS)
def test_has_rsample(jax_dist, params):
    if isinstance(jax_dist, types.FunctionType):
        pytest.skip("skip testing for non-distribution")
    jax_dist = jax_dist(**params)
    masked_dist = jax_dist.mask(False)
    indept_dist = jax_dist.expand_by([2]).to_event(1)
    transf_dist = dist.TransformedDistribution(jax_dist, biject_to(constraints.real))
    assert masked_dist.has_rsample == jax_dist.has_rsample
    assert indept_dist.has_rsample == jax_dist.has_rsample
    assert transf_dist.has_rsample == jax_dist.has_rsample

    if jax_dist.has_rsample:
        assert not jax_dist.is_discrete
        if isinstance(jax_dist, dist.TransformedDistribution):
            assert jax_dist.base_dist.has_rsample
        else:
            assert set(jax_dist.arg_constraints) == set(jax_dist.reparametrized_params)
        jax_dist.rsample(jrd.PRNGKey(0))
    else:
        with pytest.raises(NotImplementedError):
            jax_dist.rsample(jrd.PRNGKey(0))


@pytest.mark.parametrize("jax_dist, params", CONTINUOUS)
def test_sample_gradient(jax_dist, params):
    if jax_dist.__name__ in ("PowerlawPeak",):
        pytest.skip(reason=f"{jax_dist.__name__} does not provide sample method")
    if jax_dist.__name__ in ("PowerlawRedshift",):
        pytest.xfail(
            reason=f"{jax_dist.__name__} uses interpolation and is not differentiable"
        )
    if isinstance(jax_dist, types.FunctionType):
        if jax_dist.__name__ in ("NSmoothedPowerlawMSmoothedGaussian",):
            pytest.skip(reason=f"{jax_dist.__name__} does not provide sample method")
    jax_class = type(jax_dist(**params))
    reparametrized_params = [p for p in jax_class.reparametrized_params]
    if not reparametrized_params:
        pytest.skip("{} not reparametrized.".format(jax_class.__name__))

    nonrepara_params_dict = {
        k: v for k, v in params.items() if k not in reparametrized_params
    }
    repara_params = tuple(v for k, v in params.items() if k in reparametrized_params)

    rng_key = jrd.PRNGKey(0)

    def fn(args):
        args_dict = dict(zip(reparametrized_params, args))
        return jnp.sum(
            jax_dist(**args_dict, **nonrepara_params_dict).sample(key=rng_key)
        )

    actual_grad = jax.grad(fn)(repara_params)
    assert len(actual_grad) == len(repara_params)

    eps = 1e-3
    for i in range(len(repara_params)):
        if repara_params[i] is None:
            continue
        args_lhs = [p if j != i else p - eps for j, p in enumerate(repara_params)]
        args_rhs = [p if j != i else p + eps for j, p in enumerate(repara_params)]
        fn_lhs = fn(args_lhs)
        fn_rhs = fn(args_rhs)
        # finite diff approximation
        expected_grad = (fn_rhs - fn_lhs) / (2.0 * eps)
        assert jnp.shape(actual_grad[i]) == jnp.shape(repara_params[i])
        assert_allclose(jnp.sum(actual_grad[i]), expected_grad, rtol=0.02, atol=0.03)


@pytest.mark.parametrize("jax_dist, params", CONTINUOUS)
def test_jit_log_likelihood(jax_dist, params):
    if jax_dist.__name__ in (
        "EulerMaruyama",
        "GaussianRandomWalk",
        "_ImproperWrapper",
        "LKJ",
        "LKJCholesky",
        "_SparseCAR",
        "ZeroSumNormal",
        # "NPowerlawMGaussian",
    ):
        pytest.xfail(reason="non-jittable params")

    if jax_dist.__name__ in ("PowerlawPeak",):
        pytest.skip(reason=f"{jax_dist.__name__} does not provide sample method")
    if isinstance(jax_dist, types.FunctionType):
        if jax_dist.__name__ in ("NSmoothedPowerlawMSmoothedGaussian",):
            pytest.skip(reason=f"{jax_dist.__name__} does not provide sample method")
        if jax_dist.__name__ in ("NPowerlawMGaussian",):
            pytest.xfail(reason=f"{jax_dist.__name__} have shape broadcasting issues")

    rng_key = jrd.PRNGKey(0)
    samples = jax_dist(**params).sample(key=rng_key, sample_shape=(5,))

    def log_likelihood(**params):
        return jax_dist(**params).log_prob(samples)

    expected = log_likelihood(**params)
    actual = eqx.filter_jit(log_likelihood)(**params)
    assert_allclose(actual, expected, atol=2e-5, rtol=2e-5)


@pytest.mark.parametrize("jax_dist, params", CONTINUOUS)
def test_entropy_samples(jax_dist, params):
    jax_dist = jax_dist(**params)

    try:
        actual = jax_dist.entropy()
    except NotImplementedError:
        pytest.skip(reason=f"distribution {jax_dist} does not implement `entropy`")

    samples = jax_dist.sample(jrd.PRNGKey(8), (1000,))
    neg_log_probs = -jax_dist.log_prob(samples)
    mean = neg_log_probs.mean(axis=0)
    stderr = neg_log_probs.std(axis=0) / jnp.sqrt(neg_log_probs.shape[-1] - 1)
    z = (actual - mean) / stderr

    # Check the z-score is small or that all values are close. This happens, for
    # example, for uniform distributions with constant log prob and hence zero stderr.
    assert (jnp.abs(z) < 5).all() or jnp.allclose(actual, neg_log_probs, atol=1e-5)


@pytest.mark.parametrize(
    "jax_dist, params",
    # TODO: add more complete pattern for Discrete.cdf
    CONTINUOUS,
)
@pytest.mark.filterwarnings("ignore:overflow encountered:RuntimeWarning")
def test_cdf_and_icdf(jax_dist, params):
    d = jax_dist(**params)
    if d.event_dim > 0:
        pytest.skip("skip testing cdf/icdf methods of multivariate distributions")
    key1, key2 = jrd.split(jrd.PRNGKey(0))
    samples = d.sample(key=key1, sample_shape=(100,))
    quantiles = jrd.uniform(key2, (100,) + d.shape())
    try:
        atol = 1e-5
        rtol = 1e-5
        if jax_dist.__name__ in ["PowerlawRedshift"]:
            atol = 4e-3
            rtol = 0.02
        if d.shape() == () and not d.is_discrete:
            assert_allclose(
                jax.vmap(jax.grad(d.cdf))(samples),
                jnp.exp(d.log_prob(samples)),
                atol=atol,
                rtol=rtol,
            )
            assert_allclose(
                jax.vmap(jax.grad(d.icdf))(quantiles),
                jnp.exp(-d.log_prob(d.icdf(quantiles))),
                atol=atol,
                rtol=rtol,
            )
        assert_allclose(d.cdf(d.icdf(quantiles)), quantiles, atol=atol, rtol=rtol)
        assert_allclose(d.icdf(d.cdf(samples)), samples, atol=atol, rtol=rtol)
    except NotImplementedError:
        pytest.skip("cdf/icdf not implemented")


@pytest.mark.parametrize("jax_dist, params", CONTINUOUS)
def test_gof(jax_dist, params):
    if isinstance(jax_dist, types.FunctionType):
        pytest.skip("skip testing for non-distribution")
    if jax_dist.__name__ in ("PowerlawPrimaryMassRatio",):
        pytest.skip("Failure rate is lower than expected.")
    if isinstance(jax_dist, ScaledMixture):
        pytest.skip("skip testing for ScaledMixture")
    if jax_dist.__name__ in ("PowerlawRedshift",):
        pytest.skip(f"{jax_dist.__name__} is not a valid probability distribution")
    num_samples = 10000
    rng_key = jrd.PRNGKey(0)
    d = jax_dist(**params)
    try:
        samples = d.sample(key=rng_key, sample_shape=(num_samples,))
    except NotImplementedError:
        pytest.skip("sample method not implemented")
    probs = np.exp(d.log_prob(samples))

    dim = None

    # Test each batch independently.
    probs = probs.reshape(num_samples, -1)
    samples = samples.reshape(probs.shape + d.event_shape)
    for b in range(probs.shape[1]):
        try:
            gof = auto_goodness_of_fit(samples[:, b], probs[:, b], dim=dim)
        except InvalidTest:
            pytest.skip("expensive test")
        else:
            assert gof > TEST_FAILURE_RATE


@pytest.mark.parametrize("jax_dist, params", CONTINUOUS)
def test_log_prob_gradient(jax_dist, params):
    if jax_dist.__name__ in ("PowerlawPeak",):
        pytest.skip(reason=f"{jax_dist.__name__} does not provide sample method")
    if isinstance(jax_dist, types.FunctionType):
        if jax_dist.__name__ in ("NSmoothedPowerlawMSmoothedGaussian",):
            pytest.skip(reason=f"{jax_dist.__name__} does not provide sample method")
    rng_key = jrd.PRNGKey(0)
    value = jax_dist(**params).sample(rng_key)

    if isinstance(jax_dist, dist.Distribution):

        def fn(*args):
            return jnp.sum(jax_dist(*args).log_prob(value))
    else:
        params_mapping = {name: i for i, name in enumerate(params.keys())}

        def fn(*args):
            param = {k: args[params_mapping[k]] for k in params_mapping}
            return jnp.sum(jax_dist(**param).log_prob(value))

    eps = 1e-3
    for i, k in enumerate(params.keys()):
        if jax_dist is PowerlawPrimaryMassRatio and i > 1:
            continue
        if jax_dist is Wysocki2019MassModel and i != 0:
            continue
        if (jax_dist is NPowerlawMGaussian) and any(
            [k.startswith("mmin"), k.startswith("mmax"), "low" in k, "high" in k]
        ):
            continue
        if params[k] is None or jnp.result_type(params[k]) in (jnp.int32, jnp.int64):
            continue
        if isinstance(params[k], bool):
            continue
        actual_grad = jax.grad(fn, i)(*params.values())
        args_lhs = [p if j != i else p - eps for j, p in enumerate(params.values())]
        args_rhs = [p if j != i else p + eps for j, p in enumerate(params.values())]
        fn_lhs = fn(*args_lhs)
        fn_rhs = fn(*args_rhs)
        # finite diff approximation
        expected_grad = (fn_rhs - fn_lhs) / (2.0 * eps)
        assert jnp.shape(actual_grad) == jnp.shape(params[k])
        assert_allclose(jnp.sum(actual_grad), expected_grad, rtol=0.01, atol=0.01)


@pytest.mark.parametrize("jax_dist, params", CONTINUOUS)
@pytest.mark.parametrize("prepend_shape", [(), (2,), (2, 3)])
def test_distribution_constraints(jax_dist, params, prepend_shape):
    if isinstance(jax_dist, types.FunctionType):
        pytest.skip("skip testing for non-distribution")
    if isinstance(jax_dist, ScaledMixture):
        pytest.skip("skip testing for ScaledMixture")
    if jax_dist.__name__ in ("PowerlawPrimaryMassRatio",):
        pytest.skip(f"skipping test for {jax_dist.__name__}")
    valid_params = {}
    oob_params = {}
    key = jrd.PRNGKey(1)
    dependent_constraint = False
    for name, value in params.items():
        if value is None:
            oob_params[name] = None
            valid_params[name] = None
            continue
        constraint = jax_dist.arg_constraints[name]
        if isinstance(constraint, constraints._Dependent):
            dependent_constraint = True
            break
        key, key_gen = jrd.split(key)
        oob_params[name] = gen_values_outside_bounds(
            constraint, jnp.shape(value), key_gen
        )
        valid_params[name] = gen_values_within_bounds(
            constraint, jnp.shape(value), key_gen
        )

    assert jax_dist(**oob_params)

    # Invalid parameter values throw ValueError
    if not dependent_constraint:
        with pytest.raises(ValueError):
            jax_dist(**oob_params, validate_args=True)

        with pytest.raises(ValueError):
            # test error raised under jit omnistaging
            oob_params = jax.device_get(oob_params)

            def dist_gen_fn():
                d = jax_dist(**oob_params, validate_args=True)
                return d

            eqx.filter_jit(dist_gen_fn)()

    d = jax_dist(**valid_params, validate_args=True)

    # Out of support samples throw ValueError
    oob_samples = gen_values_outside_bounds(
        d.support, size=prepend_shape + d.batch_shape + d.event_shape
    )
    with pytest.warns(UserWarning, match="Out-of-support"):
        d.log_prob(oob_samples)

    # with pytest.warns(UserWarning, match="Out-of-support"):
    #     # test warning work under jit omnistaging
    #     oob_samples = jax.device_get(oob_samples)
    #     valid_params = jax.device_get(valid_params)

    #     def log_prob_fn():
    #         d = jax_dist(**valid_params, validate_args=True)
    #         return d.log_prob(oob_samples)

    #     jax.jit(log_prob_fn)()


@pytest.mark.parametrize("jax_dist, params", CONTINUOUS)
@pytest.mark.parametrize("prepend_shape", [(), (2,), (2, 3)])
@pytest.mark.parametrize("sample_shape", [(), (4,)])
def test_expand(jax_dist, params, prepend_shape, sample_shape):
    if jax_dist.__name__ in ("PowerlawPeak",):
        pytest.skip(reason=f"{jax_dist.__name__} does not provide sample method")

    if isinstance(jax_dist, types.FunctionType):
        if jax_dist.__name__ in ("NSmoothedPowerlawMSmoothedGaussian",):
            pytest.skip(reason=f"{jax_dist.__name__} does not provide sample method")
        if jax_dist.__name__ in ("NPowerlawMGaussian",):
            pytest.xfail(
                reason=f"{jax_dist.__name__} failing test cases, needs to be investigated."
            )
    jax_dist = jax_dist(**params)
    new_batch_shape = prepend_shape + jax_dist.batch_shape
    expanded_dist = jax_dist.expand(new_batch_shape)
    rng_key = jrd.PRNGKey(0)
    samples = expanded_dist.sample(rng_key, sample_shape)
    assert expanded_dist.batch_shape == new_batch_shape
    assert jnp.shape(samples) == sample_shape + new_batch_shape + jax_dist.event_shape
    assert expanded_dist.log_prob(samples).shape == sample_shape + new_batch_shape
    # test expand of expand
    assert (
        expanded_dist.expand((3,) + new_batch_shape).batch_shape
        == (3,) + new_batch_shape
    )
    # test expand error
    if prepend_shape:
        with pytest.raises(ValueError, match="Cannot broadcast distribution of shape"):
            assert expanded_dist.expand((3,) + jax_dist.batch_shape)


@pytest.mark.parametrize("jax_dist, params", CONTINUOUS)
def test_dist_pytree(jax_dist, params):
    if isinstance(jax_dist, types.FunctionType):
        pytest.skip("skip testing for non-distribution")

    def f(x):
        return jax_dist(**params)

    eqx.filter_jit(f)(0)  # this test for flatten/unflatten
    lax.map(f, np.ones(3))  # this test for compatibility w.r.t. scan
    # Test that parameters do not change after flattening.
    expected_dist = f(0)
    actual_dist = eqx.filter_jit(f)(0)
    for name in expected_dist.arg_constraints:
        expected_arg = getattr(expected_dist, name)
        actual_arg = getattr(actual_dist, name)
        assert actual_arg is not None, f"arg {name} is None"
        if np.issubdtype(np.asarray(expected_arg).dtype, np.number):
            assert_allclose(actual_arg, expected_arg)
        else:
            assert (
                actual_arg.shape == expected_arg.shape
                and actual_arg.dtype == expected_arg.dtype
            )
    try:
        expected_sample = expected_dist.sample(jrd.PRNGKey(0))
        actual_sample = actual_dist.sample(jrd.PRNGKey(0))
        expected_log_prob = expected_dist.log_prob(expected_sample)
        actual_log_prob = actual_dist.log_prob(actual_sample)
        assert_allclose(actual_sample, expected_sample, rtol=1e-6)
        assert_allclose(actual_log_prob, expected_log_prob, rtol=2e-6)
    except NotImplementedError:
        pass


def _get_vmappable_dist_init_params(jax_dist):
    if jax_dist.__name__ == ("_TruncatedCauchy"):
        return [2, 3]
    elif issubclass(jax_dist, dist.Distribution):
        init_parameters = list(inspect.signature(jax_dist.__init__).parameters.keys())[
            1:
        ]
        vmap_over_parameters = list(
            inspect.signature(vmap_over.dispatch(jax_dist)).parameters.keys()
        )[1:]
        return list(
            [
                i
                for i, name in enumerate(init_parameters)
                if name in vmap_over_parameters
            ]
        )
    else:
        raise ValueError


def _allclose_or_equal(a1, a2):
    if isinstance(a1, np.ndarray):
        return np.allclose(a2, a1)
    elif isinstance(a1, jnp.ndarray):
        return jnp.allclose(a2, a1)
    elif isinstance(a1, csr_matrix):
        return np.allclose(a2.todense(), a1.todense())
    else:
        return a2 == a1 or a2 is a1


def _tree_equal(t1, t2):
    t = jax.tree.map(_allclose_or_equal, t1, t2)
    return jnp.all(jax.flatten_util.ravel_pytree(t)[0])


@pytest.mark.parametrize("jax_dist, params", CONTINUOUS)
def test_vmap_dist(jax_dist, params):
    if isinstance(jax_dist, types.FunctionType):
        pytest.skip("skip testing for non-distribution")
    if jax_dist.__name__ in ("PowerlawRedshift",):
        pytest.xfail(f"{jax_dist.__name__} has some KeyError issues")
    param_names = list(inspect.signature(jax_dist).parameters.keys())
    vmappable_param_idxs = _get_vmappable_dist_init_params(jax_dist)
    vmappable_param_idxs = vmappable_param_idxs[: len(params)]

    if len(vmappable_param_idxs) == 0:
        return

    def make_jax_dist(**params):
        return jax_dist(**params)

    def sample(d: dist.Distribution):
        return d.sample(jrd.PRNGKey(0))

    d = make_jax_dist(**params)

    in_out_axes_cases = [
        # vmap over all args
        (
            tuple(0 if i in vmappable_param_idxs else None for i in range(len(params))),
            0,
        ),
        # vmap over a single arg, out over all attributes of a distribution
        *(
            ([0 if i == idx else None for i in range(len(params))], 0)
            for idx in vmappable_param_idxs
            if params[idx] is not None
        ),
        # vmap over a single arg, out over the associated attribute of the distribution
        *(
            (
                [0 if i == idx else None for i in range(len(params))],
                vmap_over(d, **{param_names[idx]: 0}),
            )
            for idx in vmappable_param_idxs
            if params[idx] is not None
        ),
        # vmap over a single arg, axis=1, (out single attribute, axis=1)
        *(
            (
                [1 if i == idx else None for i in range(len(params))],
                vmap_over(d, **{param_names[idx]: 1}),
            )
            for idx in vmappable_param_idxs
            if isinstance(params[idx], jnp.ndarray) and jnp.array(params[idx]).ndim > 0
        ),
    ]

    for in_axes, out_axes in in_out_axes_cases:
        batched_params = [
            (
                jax.jax.tree.map(lambda x: jnp.expand_dims(x, ax), arg)
                if isinstance(ax, int)
                else arg
            )
            for arg, ax in zip(params, in_axes)
        ]
        # Recreate the jax_dist to avoid side effects coming from `d.sample`
        # triggering lazy_property computations, which, in a few cases, break
        # vmap_over's expectations regarding existing attributes to be vmapped.
        d = make_jax_dist(*params)
        batched_d = jax.vmap(make_jax_dist, in_axes=in_axes, out_axes=out_axes)(
            *batched_params
        )
        eq = vmap(lambda x, y: _tree_equal(x, y), in_axes=(out_axes, None))(
            batched_d, d
        )
        assert eq == jnp.array([True])

        samples_dist = sample(d)
        samples_batched_dist = jax.vmap(sample, in_axes=(out_axes,))(batched_d)
        assert samples_batched_dist.shape == (1, *samples_dist.shape)



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0

from typing import Optional

import quadax
from jax import Array, numpy as jnp, random as jrd
from jax.lax import broadcast_shapes
from jax.scipy.integrate import trapezoid
from numpyro.distributions import constraints, Distribution
from numpyro.distributions.util import promote_shapes, validate_sample

from ...cosmology import PLANCK_2015_Cosmology


class PowerlawRedshift(Distribution):
    r"""Redshift distribution for compact binary mergers modeled as a power law modulated
    by the cosmological volume element.

    The probability density function is defined as:

    .. math::
        p(z) \propto \frac{dV_c/dz(z) \cdot (1 + z)^{\kappa - 1}}}, \qquad 0 \leq z \leq z_{max}

    where:
      - dV_c/dz is the differential comoving volume element,
      -  is the redshift evolution power-law index,
      - z_max is the upper redshift cutoff.

    This distribution is normalized numerically on a fixed redshift grid.

    Parameters
    ----------
    kappa : float
        The power-law exponent :math:`\kappa`.
    z_max : float
        The maximum redshift (upper limit of the support).
    """

    arg_constraints = {
        "z_max": constraints.positive,
        "kappa": constraints.real,
    }
    reparametrized_params = ["z_max", "kappa"]
    pytree_data_fields = ("_support", "kappa", "z_max")

    def __init__(
        self, z_max: Array, kappa: Array, *, validate_args: Optional[bool] = None
    ):
        self.z_max, self.kappa = promote_shapes(z_max, kappa)
        batch_shape = broadcast_shapes(jnp.shape(z_max), jnp.shape(kappa))
        self._support = constraints.interval(0.0, z_max)
        super(PowerlawRedshift, self).__init__(
            batch_shape=batch_shape, validate_args=validate_args
        )

    @constraints.dependent_property(is_discrete=False, event_dim=0)
    def support(self):
        """The support of the distribution, which is the interval [0, z_max]."""
        return self._support

    def log_differential_spacetime_volume(self, z: Array) -> Array:
        """Placeholder method for computing the differential spacetime volume."""
        logdVcdz = PLANCK_2015_Cosmology.logdVcdz(z)
        log_time_dilation = -jnp.log1p(z)
        log_differential_spacetime_volume_val = (
            log_time_dilation + logdVcdz + self.log_psi_of_z(z)
        )
        return log_differential_spacetime_volume_val

    def log_norm(self) -> Array:
        """Placeholder method for computing the log normalization constant."""
        z_grid = jnp.linspace(0.0, self.z_max, 2500)
        log_differential_spacetime_volume = self.log_differential_spacetime_volume(
            z_grid
        )
        pdfs = jnp.exp(log_differential_spacetime_volume)
        norm = trapezoid(pdfs, z_grid)
        return jnp.log(norm)

    def sample(self, key, sample_shape=()):
        """Draw samples from the distribution using inverse transform sampling.

        Note: Sampling is only supported when z_max is a static scalar.

        Parameters
        ----------
        key : jax.random.PRNGKey
            A PRNG key for sampling.
        sample_shape : tuple
            Shape of the desired sample batch.

        Returns
        -------
        samples : Array
            Redshift samples.
        """
        u = jrd.uniform(key, shape=sample_shape + self.batch_shape)
        z_grid = jnp.linspace(0.0, self.z_max, 10_000)
        pdfgrid = jnp.exp(self.log_differential_spacetime_volume(z_grid))
        norm = trapezoid(pdfgrid, z_grid)
        pdfgrid /= norm
        cdfgrid: Array = quadax.cumulative_trapezoid(pdfgrid, x=z_grid, initial=0.0)
        cdfgrid = cdfgrid / cdfgrid[-1]
        return jnp.interp(u, cdfgrid, z_grid)

    def log_psi_of_z(self, z: Array) -> Array:
        r"""Evaluate the psi function at a given redshift.

        .. math::

            \ln\psi(z) = \kappa \log(1 + z)

        Parameters
        ----------
        z : ArrayLike
            Redshift(s) to evaluate.

        Returns
        -------
        ArrayLike
            Values of the psi function.
        """
        return self.kappa * jnp.log1p(z)

    @validate_sample
    def log_prob(self, value: Array) -> Array:
        """Evaluate the log probability density function at a given redshift.

        Parameters
        ----------
        value : ArrayLike
            Redshift(s) to evaluate.

        Returns
        -------
        ArrayLike
            Log-probability values.
        """
        return self.log_differential_spacetime_volume(value)



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from . import kernel as kernel, math as math, transformations as transformations
from .kernel import log_planck_taper_window as log_planck_taper_window
from .math import (
    beta_dist_concentrations_to_mean_variance as beta_dist_concentrations_to_mean_variance,
    beta_dist_mean_variance_to_concentrations as beta_dist_mean_variance_to_concentrations,
)
from .transformations import (
    cart_to_polar as cart_to_polar,
    cart_to_spherical as cart_to_spherical,
    chi_costilt_to_chiz as chi_costilt_to_chiz,
    chieff as chieff,
    chirp_mass as chirp_mass,
    delta_m as delta_m,
    delta_m_to_symmetric_mass_ratio as delta_m_to_symmetric_mass_ratio,
    eta_from_q as eta_from_q,
    log_chirp_mass as log_chirp_mass,
    m1_m2_chi1_chi2_costilt1_costilt2_to_chieff as m1_m2_chi1_chi2_costilt1_costilt2_to_chieff,
    m1_m2_chi1_chi2_costilt1_costilt2_to_chiminus as m1_m2_chi1_chi2_costilt1_costilt2_to_chiminus,
    m1_m2_chi1z_chi2z_to_chiminus as m1_m2_chi1z_chi2z_to_chiminus,
    m1_m2_chieff_chiminus_to_chi1z_chi2z as m1_m2_chieff_chiminus_to_chi1z_chi2z,
    m1_q_to_m2 as m1_q_to_m2,
    m1_times_m2 as m1_times_m2,
    m2_q_to_m1 as m2_q_to_m1,
    m_det_z_to_m_source as m_det_z_to_m_source,
    m_source_z_to_m_det as m_source_z_to_m_det,
    mass_ratio as mass_ratio,
    Mc_eta_to_m1_m2 as Mc_eta_to_m1_m2,
    polar_to_cart as polar_to_cart,
    reduced_mass as reduced_mass,
    spherical_to_cart as spherical_to_cart,
    symmetric_mass_ratio as symmetric_mass_ratio,
    symmetric_mass_ratio_to_delta_m as symmetric_mass_ratio_to_delta_m,
    total_mass as total_mass,
)



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


import jax
from jax import numpy as jnp
from jaxtyping import ArrayLike


def m1_times_m2(m1: ArrayLike, m2: ArrayLike) -> ArrayLike:
    r"""
    .. math::
        m_1m_2(m_1, m_2) = m_1 m_2
    """
    return m1 * m2


def total_mass(m1: ArrayLike, m2: ArrayLike) -> ArrayLike:
    r"""
    .. math::
        M(m_1, m_2) = m_1 + m_2
    """
    return m1 + m2


def _mass_ratio(m1: ArrayLike, m2: ArrayLike) -> ArrayLike:
    safe_m1 = jnp.where(m1 == 0.0, 1.0, m1)
    return jnp.where(m1 == 0.0, jnp.inf, m2 / safe_m1)


def mass_ratio(m1: ArrayLike, m2: ArrayLike) -> ArrayLike:
    r"""
    .. math::
        q(m_1, m_2) = \frac{m_2}{m_1}
    """
    return jax.jit(_mass_ratio, inline=True)(m1=m1, m2=m2)


def _chirp_mass(m1: ArrayLike, m2: ArrayLike) -> ArrayLike:
    M = total_mass(m1=m1, m2=m2)
    m1m2 = m1_times_m2(m1=m1, m2=m2)
    return jnp.power(m1m2, 0.6) * jnp.power(M, -0.2)


def chirp_mass(m1: ArrayLike, m2: ArrayLike) -> ArrayLike:
    r"""
    .. math::
        M_c(m_1, m_2) = \frac{(m_1m_2)^{3/5}}{(m_1 + m_2)^{1/5}}
    """
    return jax.jit(_chirp_mass, inline=True)(m1=m1, m2=m2)


def _log_chirp_mass(m1: ArrayLike, m2: ArrayLike) -> ArrayLike:
    M = total_mass(m1=m1, m2=m2)
    return 0.6 * (jnp.log(m1) + jnp.log(m2)) - 0.2 * jnp.log(M)


def log_chirp_mass(m1: ArrayLike, m2: ArrayLike) -> ArrayLike:
    r"""
    .. math::
        \log(M_c(m_1, m_2)) = 3/5\times (\log(m_1) + \log(m_2)) - \log(m_1 + m_2)/5
    """
    return jax.jit(_log_chirp_mass, inline=True)(m1=m1, m2=m2)


def _symmetric_mass_ratio(m1: ArrayLike, m2: ArrayLike) -> ArrayLike:
    M = total_mass(m1=m1, m2=m2)
    m1m2 = m1_times_m2(m1=m1, m2=m2)
    return m1m2 * jnp.power(M, -2.0)


def symmetric_mass_ratio(m1: ArrayLike, m2: ArrayLike) -> ArrayLike:
    r"""
    .. math::
        \eta(m_1, m_2) = \frac{m_1m_2}{(m_1 + m_2)^2}
    """
    return jax.jit(_symmetric_mass_ratio, inline=True)(m1=m1, m2=m2)


def _reduced_mass(m1: ArrayLike, m2: ArrayLike) -> ArrayLike:
    M = total_mass(m1=m1, m2=m2)
    m1m2 = m1_times_m2(m1=m1, m2=m2)
    return jnp.where(M == 0.0, jnp.inf, m1m2 / M)


def reduced_mass(m1: ArrayLike, m2: ArrayLike) -> ArrayLike:
    r"""
    .. math::
        M_r(m_1, m_2) = \frac{m_1m_2}{m_1 + m_2}
    """
    return jax.jit(_reduced_mass, inline=True)(m1=m1, m2=m2)


def _delta_m(m1: ArrayLike, m2: ArrayLike) -> ArrayLike:
    diff = m1 - m2
    M = total_mass(m1=m1, m2=m2)
    return jnp.where(M == 0.0, jnp.inf, diff / M)


def delta_m(m1: ArrayLike, m2: ArrayLike) -> ArrayLike:
    r"""
    .. math::
        \delta_m(m_1, m_2) = \frac{m_1 - m_2}{m_1 + m_2}
    """
    return jax.jit(_delta_m, inline=True)(m1=m1, m2=m2)


def _delta_m_to_symmetric_mass_ratio(delta_m: ArrayLike) -> ArrayLike:
    delta_m_sq = jnp.square(delta_m)  # delta_m^2
    eta = 0.25 * (1 - delta_m_sq)  # (1 - delta_m^2) / 4
    return eta


def delta_m_to_symmetric_mass_ratio(delta_m: ArrayLike) -> ArrayLike:
    r"""
    .. math::
        \eta(\delta_m) = \frac{1 - \delta_m^2}{4}
    """
    return jax.jit(_delta_m_to_symmetric_mass_ratio, inline=True)(delta_m=delta_m)


def _symmetric_mass_ratio_to_delta_m(eta: ArrayLike) -> ArrayLike:
    eta_4 = jnp.multiply(eta, 4)  #  eta*4
    delta_m = jnp.sqrt(jnp.subtract(1, eta_4))  # sqrt(1 - 4 * eta)
    return delta_m


def symmetric_mass_ratio_to_delta_m(eta: ArrayLike) -> ArrayLike:
    r"""
    .. math::
        \delta_m(\eta) = \sqrt{1 - 4\eta}
    """
    return jax.jit(_symmetric_mass_ratio_to_delta_m, inline=True)(eta=eta)


def _m_det_z_to_m_source(m_det: ArrayLike, z: ArrayLike) -> ArrayLike:
    safe_one_more_z = jnp.where(z == -1.0, 1.0, jnp.add(1.0, z))
    m_source = jnp.where(z == -1.0, jnp.inf, m_det / safe_one_more_z)
    return m_source


def m_det_z_to_m_source(m_det: ArrayLike, z: ArrayLike) -> ArrayLike:
    r"""
    .. math::
        m_{\text{source}}(m_{\text{det}}, z) = \frac{m_{\text{det}}}{1 + z}
    """
    return jax.jit(_m_det_z_to_m_source, inline=True)(m_det=m_det, z=z)


def _m_source_z_to_m_det(m_source: ArrayLike, z: ArrayLike) -> ArrayLike:
    return m_source * (1.0 + z)


def m_source_z_to_m_det(m_source: ArrayLike, z: ArrayLike) -> ArrayLike:
    r"""
    .. math::
        m_{\text{det}}(m_{\text{source}}, z) = m_{\text{source}}(1 + z)
    """
    return jax.jit(_m_source_z_to_m_det, inline=True)(m_source=m_source, z=z)


def m1_q_to_m2(m1: ArrayLike, q: ArrayLike) -> ArrayLike:
    r"""
    .. math::
        m_2(m_1, q) = m_1q
    """
    return m1 * q


def _m2_q_to_m1(m2: ArrayLike, q: ArrayLike) -> ArrayLike:
    safe_q = jnp.where(q == 0.0, 1.0, q)
    m1 = jnp.where(q == 0.0, jnp.inf, m2 / safe_q)
    return m1


def m2_q_to_m1(m2: ArrayLike, q: ArrayLike) -> ArrayLike:
    r"""
    .. math::
        m_1(m_2, q) = \frac{m_2}{q}
    """
    return jax.jit(_m2_q_to_m1, inline=True)(m2=m2, q=q)


def chi_costilt_to_chiz(chi: ArrayLike, costilt: ArrayLike) -> ArrayLike:
    r"""
    .. math::
        \chi_z(\chi, \cos(\theta)) = \chi \cos(\theta)
    """
    return chi * costilt


def _m1_m2_chi1z_chi2z_to_chiminus(
    m1: ArrayLike, m2: ArrayLike, chi1z: ArrayLike, chi2z: ArrayLike
) -> ArrayLike:
    m1_chi1z = m1 * chi1z
    m2_chi2z = m2 * chi2z
    M = total_mass(m1=m1, m2=m2)
    diff = m1_chi1z - m2_chi2z
    return jnp.where(M == 0, jnp.inf, diff / M)


def m1_m2_chi1z_chi2z_to_chiminus(
    m1: ArrayLike, m2: ArrayLike, chi1z: ArrayLike, chi2z: ArrayLike
) -> ArrayLike:
    r"""
    .. math::
        \chi_{\text{minus}}(m_1, m_2, \chi_{1z}, \chi_{2z}) = \frac{m_1\chi_{1z} - m_2\chi_{2z}}{m_1 + m_2}
    """
    return jax.jit(_m1_m2_chi1z_chi2z_to_chiminus, inline=True)(
        m1=m1, m2=m2, chi1z=chi1z, chi2z=chi2z
    )


def _chieff(
    m1: ArrayLike, m2: ArrayLike, chi1z: ArrayLike, chi2z: ArrayLike
) -> ArrayLike:
    m1_chi1z = m1 * chi1z
    m2_chi2z = m2 * chi2z
    M = total_mass(m1=m1, m2=m2)
    m_dot_chi = m1_chi1z + m2_chi2z
    return jnp.where(M == 0, jnp.inf, m_dot_chi / M)


def chieff(
    m1: ArrayLike, m2: ArrayLike, chi1z: ArrayLike, chi2z: ArrayLike
) -> ArrayLike:
    r"""
    .. math::
        \chi_{\text{eff}}(m_1, m_2, \chi_{1z}, \chi_{2z}) = \frac{m_1\chi_{1z} + m_2\chi_{2z}}{m_1 + m_2}
    """
    return jax.jit(_chieff, inline=True)(m1=m1, m2=m2, chi1z=chi1z, chi2z=chi2z)


def m1_m2_chi1_chi2_costilt1_costilt2_to_chieff(
    *,
    m1: ArrayLike,
    m2: ArrayLike,
    chi1: ArrayLike,
    chi2: ArrayLike,
    costilt1: ArrayLike,
    costilt2: ArrayLike,
) -> ArrayLike:
    r"""
    .. math::
        \chi_{\text{eff}}(m_1, m_2, \chi_1, \chi_2, \cos(\theta_1), \cos(\theta_2)) =
        \frac{m_1\chi_1\cos(\theta_1) + m_2\chi_2\cos(\theta_2)}{m_1 + m_2}
    """
    chi1z = chi_costilt_to_chiz(chi=chi1, costilt=costilt1)
    chi2z = chi_costilt_to_chiz(chi=chi2, costilt=costilt2)
    return chieff(m1=m1, m2=m2, chi1z=chi1z, chi2z=chi2z)


def m1_m2_chi1_chi2_costilt1_costilt2_to_chiminus(
    *,
    m1: ArrayLike,
    m2: ArrayLike,
    chi1: ArrayLike,
    chi2: ArrayLike,
    costilt1: ArrayLike,
    costilt2: ArrayLike,
) -> ArrayLike:
    r"""
    .. math::
        \chi_{\text{minus}}(m_1, m_2, \chi_1, \chi_2, \cos(\theta_1), \cos(\theta_2)) =
        \frac{m_1\chi_1\cos(\theta_1) - m_2\chi_2\cos(\theta_2)}{m_1 + m_2}
    """
    chi1z = chi_costilt_to_chiz(chi=chi1, costilt=costilt1)
    chi2z = chi_costilt_to_chiz(chi=chi2, costilt=costilt2)
    return m1_m2_chi1z_chi2z_to_chiminus(m1=m1, m2=m2, chi1z=chi1z, chi2z=chi2z)


def m1_m2_chieff_chiminus_to_chi1z_chi2z(
    m1: ArrayLike, m2: ArrayLike, chieff: ArrayLike, chiminus: ArrayLike
) -> tuple[ArrayLike, ArrayLike]:
    r"""
    .. math::
        \begin{align*}
        \chi_{1z}(m_1, m_2, \chi_{\text{eff}}, \chi_{\text{minus}}) &=
        \frac{m_1+m_2}{2m_1} \left( \chi_{\text{eff}} + \chi_{\text{minus}} \right)\\
        \chi_{2z}(m_1, m_2, \chi_{\text{eff}}, \chi_{\text{minus}}) &=
        \frac{m_1+m_2}{2m_2} \left( \chi_{\text{eff}} - \chi_{\text{minus}} \right)
        \end{align*}
    """
    half_M = jnp.multiply(0.5, total_mass(m1=m1, m2=m2))  # M/2
    chi1z = jnp.divide(
        jnp.multiply(half_M, jnp.add(chieff, chiminus)), m1
    )  # chi1z = M/2 * (chieff + chiminus) / m1
    chi2z = jnp.divide(
        jnp.multiply(half_M, jnp.subtract(chieff, chiminus)), m2
    )  # chi2z = M/2 * (chieff - chiminus) / m2
    return chi1z, chi2z


def Mc_eta_to_m1_m2(Mc: ArrayLike, eta: ArrayLike) -> tuple[ArrayLike, ArrayLike]:
    r"""
    .. math::
        \begin{align*}
            m_1(M_c, \eta) &= \frac{M_c}{2} \eta^{-0.6} (1 + \sqrt{1 - 4\eta}) \\
            m_2(M_c, \eta) &= \frac{M_c}{2} \eta^{-0.6} (1 - \sqrt{1 - 4\eta})
        \end{align*}
    """
    delta_sq = jnp.subtract(1, jnp.multiply(4.0, eta))  # 1 - 4 * eta
    delta_sq = jnp.maximum(
        delta_sq, jnp.zeros_like(delta_sq)
    )  # to avoid negative values
    delta = jnp.sqrt(delta_sq)  # sqrt(1 - 4 * eta)
    half_Mc = jnp.multiply(0.5, Mc)  # Mc/2
    eta_pow_neg_point_six = jnp.power(eta, -0.6)  # eta^-0.6
    half_Mc_times_eta_pow_neg_point_six = jnp.multiply(
        half_Mc, eta_pow_neg_point_six
    )  # Mc/2 * eta^-0.6
    m2 = jnp.multiply(
        half_Mc_times_eta_pow_neg_point_six, jnp.subtract(1.0, delta)
    )  # m2 = Mc/2 * eta^-0.6 * (1 - delta)
    m1 = jnp.multiply(
        half_Mc_times_eta_pow_neg_point_six, jnp.add(1.0, delta)
    )  # m1 = Mc/2 * eta^-0.6 * (1 + delta)
    return m1, m2


def _eta_from_q(q: ArrayLike) -> ArrayLike:
    safe_q_from_neg = jnp.where(q <= 0, 1, q)
    log_eta = jnp.where(
        q <= 0, -jnp.inf, jnp.log(safe_q_from_neg) - 2.0 * jnp.log1p(safe_q_from_neg)
    )
    return jnp.exp(log_eta)


def eta_from_q(q: ArrayLike) -> ArrayLike:
    r"""
    .. math::
        \eta(q) = \frac{q}{(1 + q)^2}
    """
    return jax.jit(_eta_from_q, inline=True)(q)


def polar_to_cart(r: ArrayLike, theta: ArrayLike) -> tuple[ArrayLike, ArrayLike]:
    r"""
    .. math::
        \begin{align*}
            x(r, \theta) &= r \cos(\theta) \\
            y(r, \theta) &= r \sin(\theta)
        \end{align*}
    """
    x = r * jnp.cos(theta)
    y = r * jnp.sin(theta)
    return x, y


def cart_to_polar(x: ArrayLike, y: ArrayLike) -> tuple[ArrayLike, ArrayLike]:
    r"""
    .. math::
        \begin{align*}
            r(x, y) &= \sqrt{x^2 + y^2} \\
            \theta(x, y) &= \arctan(y/x)
        \end{align*}
    """
    r = jnp.sqrt(x * x + y * y)
    theta = jnp.arctan2(y, x)
    return r, theta


def spherical_to_cart(
    r: ArrayLike, theta: ArrayLike, phi: ArrayLike
) -> tuple[ArrayLike, ArrayLike, ArrayLike]:
    r"""
    .. math::
        \begin{align*}
            x(r, \theta, \phi) &= r \sin(\theta) \cos(\phi) \\
            y(r, \theta, \phi) &= r \sin(\theta) \sin(\phi) \\
            z(r, \theta, \phi) &= r \cos(\theta)
        \end{align*}
    """
    x = r * jnp.sin(theta) * jnp.cos(phi)  # x = r * sin(theta) * cos(phi)
    y = r * jnp.sin(theta) * jnp.sin(phi)  # y = r * sin(theta) * sin(phi)
    z = r * jnp.cos(theta)  # z = r * cos(theta)
    return x, y, z


def cart_to_spherical(
    x: ArrayLike, y: ArrayLike, z: ArrayLike
) -> tuple[ArrayLike, ArrayLike, ArrayLike]:
    r"""
    .. math::
        \begin{align*}
            r(x, y, z) &= \sqrt{x^2 + y^2 + z^2} \\
            \theta(x, y, z) &= \arccos\left(\frac{z}{r}\right) \\
            \phi(x, y, z) &= \arctan\left(\frac{y}{x}\right)
        \end{align*}
    """
    r = jnp.sqrt(x * x + y * y + z * z)  # r = sqrt(x^2 + y^2 + z^2)
    safe_r = jnp.where(r == 0.0, 1.0, r)
    theta = jnp.arccos(
        jnp.where(r == 0.0, jnp.inf, z / safe_r)
    )  # theta = arccos(z / r)
    phi = jnp.arctan2(y, x)  # phi = arctan(y / x)
    return r, theta, phi



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from typing import Union

from jax import numpy as jnp
from jaxtyping import Array


def check_min_concentration_for_beta_dist(
    loc: Array,
    var: Array,
    /,
    *,
    alpha_min: Union[Array, float] = 1.0,
    beta_min: Union[Array, float] = 1.0,
) -> Array:
    r"""Check if the mean and variance are valid for a given minimum alpha and beta of a
    Beta distribution.

    .. math::
        \alpha > \alpha_{\mathrm{min}} \iff \mu^2 (1 - \mu) > \sigma^2 (\alpha_{\mathrm{min}} + \mu)

    .. math::
        \beta > \beta_{\mathrm{min}} \iff \mu (1 - \mu)^2 > \sigma^2 (\beta_{\mathrm{min}} + 1 - \mu)

    Parameters
    ----------
    loc : Array
        The location parameter (mean).
    var : Array
        The variance parameter.
    alpha_min : Union[Array, float]
        The minimum allowed value for alpha, default is 1.0.
    beta_min : Union[Array, float]
        The minimum allowed value for beta, default is 1.0.

    Returns
    -------
    Array
        A boolean array indicating whether the constraints are satisfied.
    """
    ensure_positive_alpha = jnp.greater(
        jnp.square(loc) * (1.0 - loc),
        var * (alpha_min + loc),
    )
    ensure_positive_beta = jnp.greater(
        loc * jnp.square(1.0 - loc),
        var * (beta_min + 1.0 - loc),
    )
    valid_var = jnp.logical_and(ensure_positive_alpha, ensure_positive_beta)
    return valid_var



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


#
"""The population analysis and estimation of merger rates of compact binaries is one of
the important topics in gravitational wave astronomy.

The primary ingredient in these analyses is the population-averaged sensitive volume.
"""

from ._abc import VolumeTimeSensitivityInterface as VolumeTimeSensitivityInterface
from ._neuralpdet import (
    NeuralNetProbabilityOfDetection as NeuralNetProbabilityOfDetection,
)
from ._neuralvt import NeuralNetVolumeTimeSensitivity as NeuralNetVolumeTimeSensitivity
from ._pdet import pdet_O3 as pdet_O3
from ._popmodelvt import (
    PopModelsCalibratedVolumeTimeSensitivity as PopModelsCalibratedVolumeTimeSensitivity,
    PopModelsVolumeTimeSensitivity as PopModelsVolumeTimeSensitivity,
)
from ._semianalyticalinjvt import (
    SemiAnalyticalRealInjectionVolumeTimeSensitivity as SemiAnalyticalRealInjectionVolumeTimeSensitivity,
)
from ._syninjvt import (
    SyntheticInjectionVolumeTimeSensitivity as SyntheticInjectionVolumeTimeSensitivity,
)
from ._train import train_regressor as train_regressor
from ._utils import (
    load_model as load_model,
    make_model as make_model,
    mse_loss_fn as mse_loss_fn,
    predict as predict,
    read_data as read_data,
    save_model as save_model,
)
from ._vt import available as available



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


def ensure_dat_extension(filename: str) -> str:
    """Transform a filename to end with .dat if it does not have an extension.

    Parameters
    ----------
    filename : str
        Name of the file

    Returns
    -------
    str
        Filename ending with .dat

    Raises
    ------
    ValueError
        If filename has an extension other than .dat
    """
    if filename.endswith(".dat"):
        return filename
    elif "." not in filename:
        return filename + ".dat"
    else:
        ext = filename.split(".")[-1]
        raise ValueError(
            f"Invalid filename {filename!r}: found extension '.{ext}' but must end with '.dat' or have no extension"
        )



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0

import argparse
import glob
import os

import arviz as az
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt


def make_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="Command line interface for plotting Effective Sample Size per draw.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--train-chain-regex",
        help="regex pattern for the train chains files. Only .dat files are supported.",
        required=True,
        type=str,
    )
    parser.add_argument(
        "--production-chain-regex",
        help="regex pattern for the production chains files. Only .dat files are supported.",
        required=True,
        type=str,
    )
    parser.add_argument(
        "-o",
        "--output",
        help="output file path",
        required=True,
        type=argparse.FileType("w"),
    )
    parser.add_argument(
        "--labels",
        nargs="+",
        help="labels for the chains",
        required=True,
    )
    parser.add_argument(
        "--width",
        help="width of the plot in inches",
        default=12,
        type=int,
    )
    parser.add_argument(
        "--height",
        help="height of the plot in inches",
        default=6,
        type=int,
    )
    parser.add_argument(
        "--use-latex",
        help="use LaTeX for rendering text",
        action="store_true",
    )
    parser.add_argument(
        "--font-family",
        help="font family to use",
        type=str,
        default=None,
    )
    parser.add_argument(
        "--dpi",
        help="dots per inch to save file",
        type=int,
        default=100,
    )

    return parser


def main() -> None:
    """Main function of the script."""
    parser = make_parser()
    args = parser.parse_args()

    plt.rcParams.update({"text.usetex": args.use_latex})
    if args.font_family is not None:
        plt.rcParams.update({"font.family": args.font_family})

    train_paths = glob.glob(args.train_chain_regex)
    prod_paths = glob.glob(args.production_chain_regex)

    train_chains = []

    for file_path in train_paths:
        chain = np.loadtxt(file_path)
        train_chains.append(chain)

    train_chains = np.array(train_chains)

    train_data = az.convert_to_inference_data(train_chains)

    prod_chains = []

    for file_path in prod_paths:
        chain = np.loadtxt(file_path)
        prod_chains.append(chain)

    prod_chains = np.array(prod_chains)

    prod_data = az.convert_to_inference_data(prod_chains)

    ess_train = az.ess(train_data)
    ess_prod = az.ess(prod_data)

    n_draws = train_data.posterior.dims["draw"]

    sns.set_theme(style="whitegrid")
    labels = args.labels
    df = pd.DataFrame(
        data={
            "Parameter": labels,
            "Training": ess_train.to_array().data.squeeze() / n_draws,
            "Production": ess_prod.to_array().data.squeeze() / n_draws,
        }
    )

    height, width = args.height, args.width
    aspect = width / height

    ax = sns.catplot(
        data=df.melt(id_vars="Parameter"),
        x="Parameter",
        y="value",
        hue="variable",
        kind="bar",
        height=height,
        aspect=aspect,
        margin_titles=True,
    )
    ax.set_xticklabels(rotation=45)
    ax.set_axis_labels("Parameter", "ESS per draw")
    ax.legend.draw_frame(True)
    ax.legend.set_title("")
    ax.legend.set_bbox_to_anchor((1, 0.9))

    sub_plot = ax.facet_axis(0, 0)

    for c in sub_plot.containers:
        labels = [f"{h.get_height():.3f}" for h in c]
        sub_plot.bar_label(c, labels=labels, label_type="edge")

    plt.tight_layout()

    output_ext = os.path.splitext(args.output.name)[1].lower()
    plt_savefig_kwargs = dict()
    if output_ext == ".png":
        plt_savefig_kwargs["dpi"] = args.dpi
    plt.savefig(args.output.name, bbox_inches="tight", **plt_savefig_kwargs)
    plt.close("all")



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from collections.abc import Callable, Sequence
from typing import Dict, Optional, Union

import equinox as eqx
import h5py
import jax
from jax import lax, numpy as jnp
from jaxtyping import Array

from ._utils import load_model


class NeuralNetProbabilityOfDetection(eqx.Module):
    shuffle_indices: Optional[Sequence[int]] = eqx.field(static=True, default=None)
    """The indices to shuffle the input to the model."""
    batch_size: Optional[int] = eqx.field(static=True, default=None)
    """The batch size used by :func:`jax.lax.map` in mapped functions."""
    neural_vt_model: eqx.nn.MLP
    """The neural volume-time sensitivity model."""
    parameter_ranges: Optional[Dict[str, Union[int, float]]]
    """Ranges of the parameters expected by the model."""

    def __init__(
        self,
        parameters: Sequence[str],
        filename: str,
        batch_size: Optional[int] = None,
    ) -> None:
        """Convenience class for loading a neural pdet.

        Parameters
        ----------
        parameters : Sequence[str]
            The names of the parameters that the model expects.
        filename : str
            The filename of the neural vt.
        batch_size : Optional[int], optional
            The batch size :func:`jax.lax.map` should use, by default None.
        """
        if not parameters:
            raise ValueError("parameters sequence cannot be empty")
        if not isinstance(parameters, Sequence):
            raise TypeError(f"parameters must be a Sequence, got {type(parameters)}")
        if not all(isinstance(p, str) for p in parameters):
            raise TypeError("all parameters must be strings")
        if batch_size is not None:
            if not isinstance(batch_size, int):
                raise TypeError(
                    f"batch_size must be an integer, got {type(batch_size)}"
                )
            if batch_size < 1:
                raise ValueError(
                    f"batch_size must be a positive integer, got {batch_size}"
                )

        self.batch_size = batch_size
        names, self.neural_vt_model = load_model(filename)
        if any(name not in parameters for name in names):
            raise ValueError(
                f"Model in {filename} expects parameters {names}, but received "
                f"{parameters}. Missing: {set(names) - set(parameters)}"
            )
        self.shuffle_indices = [parameters.index(name) for name in names]
        parameter_ranges = {}
        with h5py.File(filename, "r") as f:
            for k, v in f.attrs.items():
                if k.endswith("_min") or k.endswith("_max"):
                    parameter_ranges[k] = lax.stop_gradient(v)
        if len(parameter_ranges) > 0:
            self.parameter_ranges = parameter_ranges
        else:
            self.parameter_ranges = None

    def get_logVT(self) -> Callable[[Array], Array]:  # TODO: rename to get_pdet
        """Gets the logVT function."""

        @jax.jit
        def _pdet(x: Array) -> Array:
            x_new = x[..., self.shuffle_indices]
            y_new = self.neural_vt_model(x_new)
            mask = jnp.less_equal(y_new, 0.0)
            safe_y_new = jnp.where(mask, 1.0, y_new)
            return jnp.where(mask, -jnp.inf, jnp.log(safe_y_new))

        return _pdet

    def get_mapped_logVT(
        self,
    ) -> Callable[[Array], Array]:  # TODO: rename to get_mapped_pdet
        """Gets the vmapped logVT function for batch processing."""
        _batch_size = self.batch_size

        @jax.jit
        def _pdet(x: Array) -> Array:
            x_new = x[..., self.shuffle_indices]
            y_new = jnp.squeeze(
                lax.map(self.neural_vt_model, x_new, batch_size=_batch_size),
                axis=-1,
            )
            mask = jnp.less_equal(y_new, 0.0)
            safe_y_new = jnp.where(mask, 1.0, y_new)
            return jnp.where(mask, -jnp.inf, jnp.log(safe_y_new))

        return _pdet



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


import json
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from functools import partial
from typing import List, Tuple

from jax import numpy as jnp, random as jrd
from numpyro import distributions as dist

import gwkokab
from gwkokab.errors import banana_error_m1_m2, truncated_normal_error
from gwkokab.models import NPowerlawMGaussian
from gwkokab.models.npowerlawmgaussian._ncombination import (
    create_truncated_normal_distributions,
)
from gwkokab.parameters import Parameters
from gwkokab.poisson_mean import PoissonMean
from gwkokab.population import error_magazine, PopulationFactory
from kokab.utils import genie_parser, poisson_mean_parser
from kokab.utils.common import expand_arguments, vt_json_read_and_process
from kokab.utils.regex import match_all


def make_parser() -> ArgumentParser:
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    parser = genie_parser.get_parser(parser)

    model_group = parser.add_argument_group("Model Options")

    model_group.add_argument(
        "--model-json",
        help="Path to the JSON file containing the model parameters",
        type=str,
        required=True,
    )

    spin_group = model_group.add_mutually_exclusive_group()
    spin_group.add_argument(
        "--add-beta-spin",
        action="store_true",
        help="Include beta spin parameters in the model.",
    )
    spin_group.add_argument(
        "--add-truncated-normal-spin",
        action="store_true",
        help="Include truncated normal spin parameters in the model.",
    )

    model_group.add_argument(
        "--add-tilt",
        action="store_true",
        help="Include tilt parameters in the model.",
    )
    model_group.add_argument(
        "--add-truncated-normal-eccentricity",
        action="store_true",
        help="Include truncated normal eccentricity parameters in the model.",
    )
    model_group.add_argument(
        "--add-mean-anomaly",
        action="store_true",
        help="Include mean_anomaly parameter in the model",
    )
    model_group.add_argument(
        "--add-redshift",
        action="store_true",
        help="Include redshift parameter in the model",
    )
    model_group.add_argument(
        "--add-cos-iota",
        action="store_true",
        help="Include cos_iota parameter in the model",
    )
    model_group.add_argument(
        "--add-polarization-angle",
        action="store_true",
        help="Include polarization_angle parameter in the model",
    )
    model_group.add_argument(
        "--add-right-ascension",
        action="store_true",
        help="Include right_ascension parameter in the model",
    )
    model_group.add_argument(
        "--add-sin-declination",
        action="store_true",
        help="Include sin_declination parameter in the model",
    )
    model_group.add_argument(
        "--add-detection-time",
        action="store_true",
        help="Include detection_time parameter in the model",
    )
    model_group.add_argument(
        "--add-phi-1",
        action="store_true",
        help="Include phi_1 parameter in the model",
    )
    model_group.add_argument(
        "--add-phi-2",
        action="store_true",
        help="Include phi_2 parameter in the model",
    )
    model_group.add_argument(
        "--add-phi-12",
        action="store_true",
        help="Include phi_12 parameter in the model",
    )
    model_group.add_argument(
        "--add-phi-orb",
        action="store_true",
        help="Include phi_orb parameter in the model",
    )

    err_group = parser.add_argument_group("Error Options")
    err_group.add_argument(
        "--err-json",
        help="Path to the JSON file containing the error parameters",
        type=str,
        required=True,
    )

    return parser


def main() -> None:
    """Main function of the script."""
    parser = make_parser()
    args = parser.parse_args()

    with open(args.model_json, "r") as f:
        model_json = json.load(f)

    with open(args.err_json, "r") as f:
        err_json = json.load(f)

    N_pl = model_json["N_pl"]
    N_g = model_json["N_g"]

    has_spin = args.add_beta_spin or args.add_truncated_normal_spin
    has_tilt = args.add_tilt
    has_eccentricity = args.add_truncated_normal_eccentricity
    has_mean_anomaly = args.add_mean_anomaly
    has_redshift = args.add_redshift
    has_cos_iota = args.add_cos_iota
    has_polarization_angle = args.add_polarization_angle
    has_right_ascension = args.add_right_ascension
    has_sin_declination = args.add_sin_declination
    has_detection_time = args.add_detection_time
    has_phi_1 = args.add_phi_1
    has_phi_2 = args.add_phi_2
    has_phi_12 = args.add_phi_12
    has_phi_orb = args.add_phi_orb

    err_params_name = ["scale_eta", "scale_Mc"]
    if has_spin:
        err_params_name.extend(
            [
                "chi1_high",
                "chi1_low",
                "chi1_scale",
                "chi2_high",
                "chi2_low",
                "chi2_scale",
            ]
        )
    if has_tilt:
        err_params_name.extend(
            [
                "cos_tilt_1_high",
                "cos_tilt_1_low",
                "cos_tilt_1_scale",
                "cos_tilt_2_high",
                "cos_tilt_2_low",
                "cos_tilt_2_scale",
            ]
        )
    if has_eccentricity:
        err_params_name.extend(["ecc_high", "ecc_low", "ecc_scale"])
    if has_mean_anomaly:
        err_params_name.extend(
            [
                Parameters.MEAN_ANOMALY.value + "_high",
                Parameters.MEAN_ANOMALY.value + "_low",
                Parameters.MEAN_ANOMALY.value + "_scale",
            ]
        )
    if has_redshift:
        err_params_name.extend(["redshift_high", "redshift_low", "redshift_scale"])
    if has_cos_iota:
        err_params_name.extend(
            [
                Parameters.COS_IOTA.value + "_high",
                Parameters.COS_IOTA.value + "_low",
                Parameters.COS_IOTA.value + "_scale",
            ]
        )
    if has_polarization_angle:
        err_params_name.extend(
            [
                Parameters.POLARIZATION_ANGLE.value + "_high",
                Parameters.POLARIZATION_ANGLE.value + "_low",
                Parameters.POLARIZATION_ANGLE.value + "_scale",
            ]
        )
    if has_right_ascension:
        err_params_name.extend(
            [
                Parameters.RIGHT_ASCENSION.value + "_high",
                Parameters.RIGHT_ASCENSION.value + "_low",
                Parameters.RIGHT_ASCENSION.value + "_scale",
            ]
        )
    if has_sin_declination:
        err_params_name.extend(
            [
                Parameters.SIN_DECLINATION.value + "_high",
                Parameters.SIN_DECLINATION.value + "_low",
                Parameters.SIN_DECLINATION.value + "_scale",
            ]
        )
    if has_detection_time:
        err_params_name.append(Parameters.DETECTION_TIME.value + "_scale")
    if has_phi_1:
        err_params_name.extend(
            [
                Parameters.PHI_1.value + "_high",
                Parameters.PHI_1.value + "_low",
                Parameters.PHI_1.value + "_scale",
            ]
        )
    if has_phi_2:
        err_params_name.extend(
            [
                Parameters.PHI_2.value + "_high",
                Parameters.PHI_2.value + "_low",
                Parameters.PHI_2.value + "_scale",
            ]
        )
    if has_phi_12:
        err_params_name.extend(
            [
                Parameters.PHI_12.value + "_high",
                Parameters.PHI_12.value + "_low",
                Parameters.PHI_12.value + "_scale",
            ]
        )
    if has_phi_orb:
        err_params_name.extend(
            [
                Parameters.PHI_ORB.value + "_high",
                Parameters.PHI_ORB.value + "_low",
                Parameters.PHI_ORB.value + "_scale",
            ]
        )

    err_params_value = match_all(err_params_name, err_json)

    all_params: List[Tuple[str, int]] = [
        ("alpha_pl", N_pl),
        ("beta_pl", N_pl),
        ("log_rate", N_pl + N_g),
        ("m1_high_g", N_g),
        ("m1_loc_g", N_g),
        ("m1_low_g", N_g),
        ("m1_scale_g", N_g),
        ("m2_high_g", N_g),
        ("m2_loc_g", N_g),
        ("m2_low_g", N_g),
        ("m2_scale_g", N_g),
        ("mmax_pl", N_pl),
        ("mmin_pl", N_pl),
    ]

    parameters_name: Tuple[str, ...] = (
        Parameters.PRIMARY_MASS_SOURCE.value,
        Parameters.SECONDARY_MASS_SOURCE.value,
    )

    error_magazine.register(
        (Parameters.PRIMARY_MASS_SOURCE.value, Parameters.SECONDARY_MASS_SOURCE.value),
        lambda x, size, key: banana_error_m1_m2(
            x,
            size,
            key,
            scale_Mc=err_params_value["scale_Mc"],
            scale_eta=err_params_value["scale_eta"],
        ),
    )

    if has_spin:
        parameters_name += (
            Parameters.PRIMARY_SPIN_MAGNITUDE.value,
            Parameters.SECONDARY_SPIN_MAGNITUDE.value,
        )
        if args.add_truncated_normal_spin:
            gwkokab.models.npowerlawmgaussian._model.build_spin_distributions = (
                create_truncated_normal_distributions
            )
            all_params.extend(
                [
                    ("chi1_high_g", N_g),
                    ("chi1_high_pl", N_pl),
                    ("chi1_loc_g", N_g),
                    ("chi1_loc_pl", N_pl),
                    ("chi1_low_g", N_g),
                    ("chi1_low_pl", N_pl),
                    ("chi1_scale_g", N_g),
                    ("chi1_scale_pl", N_pl),
                    ("chi2_high_g", N_g),
                    ("chi2_high_pl", N_pl),
                    ("chi2_loc_g", N_g),
                    ("chi2_loc_pl", N_pl),
                    ("chi2_low_g", N_g),
                    ("chi2_low_pl", N_pl),
                    ("chi2_scale_g", N_g),
                    ("chi2_scale_pl", N_pl),
                ]
            )
        if args.add_beta_spin:
            all_params.extend(
                [
                    ("chi1_mean_g", N_g),
                    ("chi1_mean_pl", N_pl),
                    ("chi1_variance_g", N_g),
                    ("chi1_variance_pl", N_pl),
                    ("chi2_mean_g", N_g),
                    ("chi2_mean_pl", N_pl),
                    ("chi2_variance_g", N_g),
                    ("chi2_variance_pl", N_pl),
                ]
            )

        error_magazine.register(
            Parameters.PRIMARY_SPIN_MAGNITUDE.value,
            partial(
                truncated_normal_error,
                scale=err_params_value["chi1_scale"],
                low=err_params_value.get("chi1_low"),
                high=err_params_value.get("chi1_high"),
                cut_low=-1.0,
                cut_high=1.0,
            ),
        )

        error_magazine.register(
            Parameters.SECONDARY_SPIN_MAGNITUDE.value,
            partial(
                truncated_normal_error,
                scale=err_params_value["chi2_scale"],
                low=err_params_value.get("chi2_low"),
                high=err_params_value.get("chi2_high"),
                cut_low=-1.0,
                cut_high=1.0,
            ),
        )

    if has_tilt:
        parameters_name += (Parameters.COS_TILT_1.value, Parameters.COS_TILT_2.value)
        all_params.extend(
            [
                ("cos_tilt_zeta_g", N_g),
                ("cos_tilt_zeta_pl", N_pl),
                ("cos_tilt1_scale_g", N_g),
                ("cos_tilt1_scale_pl", N_pl),
                ("cos_tilt2_scale_g", N_g),
                ("cos_tilt2_scale_pl", N_pl),
            ]
        )

        error_magazine.register(
            Parameters.COS_TILT_1.value,
            partial(
                truncated_normal_error,
                scale=err_params_value["cos_tilt_1_scale"],
                low=err_params_value.get("cos_tilt_1_low"),
                high=err_params_value.get("cos_tilt_1_high"),
                cut_low=-1.0,
                cut_high=1.0,
            ),
        )

        error_magazine.register(
            Parameters.COS_TILT_2.value,
            partial(
                truncated_normal_error,
                scale=err_params_value["cos_tilt_2_scale"],
                low=err_params_value.get("cos_tilt_2_low"),
                high=err_params_value.get("cos_tilt_2_high"),
                cut_low=-1.0,
                cut_high=1.0,
            ),
        )

    if has_phi_1:
        parameters_name += (Parameters.PHI_1.value,)

        all_params.extend(
            [
                (Parameters.PHI_1.value + "_high_g", N_g),
                (Parameters.PHI_1.value + "_high_pl", N_pl),
                (Parameters.PHI_1.value + "_low_g", N_g),
                (Parameters.PHI_1.value + "_low_pl", N_pl),
            ]
        )

        @error_magazine.register(Parameters.PHI_1.value)
        def phi_1_error(x, size, key):
            err_x = dist.TruncatedNormal(
                loc=x,
                scale=err_params_value[Parameters.PHI_1.value + "_scale"],
                low=err_params_value.get(Parameters.PHI_1.value + "_low", 0.0),
                high=err_params_value.get(Parameters.PHI_1.value + "_high", 2 * jnp.pi),
            ).sample(key=key, sample_shape=(size,))

            err_x = jnp.mod(err_x, 2 * jnp.pi)
            return err_x

    if has_phi_2:
        parameters_name += (Parameters.PHI_2.value,)

        all_params.extend(
            [
                (Parameters.PHI_2.value + "_high_g", N_g),
                (Parameters.PHI_2.value + "_high_pl", N_pl),
                (Parameters.PHI_2.value + "_low_g", N_g),
                (Parameters.PHI_2.value + "_low_pl", N_pl),
            ]
        )

        @error_magazine.register(Parameters.PHI_2.value)
        def phi_2_error(x, size, key):
            err_x = dist.TruncatedNormal(
                loc=x,
                scale=err_params_value[Parameters.PHI_2.value + "_scale"],
                low=err_params_value.get(Parameters.PHI_2.value + "_low", 0.0),
                high=err_params_value.get(Parameters.PHI_2.value + "_high", 2 * jnp.pi),
            ).sample(key=key, sample_shape=(size,))

            err_x = jnp.mod(err_x, 2 * jnp.pi)
            return err_x

    if has_phi_12:
        parameters_name += (Parameters.PHI_12.value,)

        all_params.extend(
            [
                (Parameters.PHI_12.value + "_high_g", N_g),
                (Parameters.PHI_12.value + "_high_pl", N_pl),
                (Parameters.PHI_12.value + "_low_g", N_g),
                (Parameters.PHI_12.value + "_low_pl", N_pl),
            ]
        )

        @error_magazine.register(Parameters.PHI_12.value)
        def phi_12_error(x, size, key):
            err_x = dist.TruncatedNormal(
                loc=x,
                scale=err_params_value[Parameters.PHI_12.value + "_scale"],
                low=err_params_value.get(Parameters.PHI_12.value + "_low", 0.0),
                high=err_params_value.get(
                    Parameters.PHI_12.value + "_high", 2 * jnp.pi
                ),
            ).sample(key=key, sample_shape=(size,))

            err_x = jnp.mod(err_x, 2 * jnp.pi)
            return err_x

    if has_eccentricity:
        parameters_name += (Parameters.ECCENTRICITY.value,)
        all_params.extend(
            [
                ("ecc_high_g", N_g),
                ("ecc_high_pl", N_pl),
                ("ecc_loc_g", N_g),
                ("ecc_loc_pl", N_pl),
                ("ecc_low_g", N_g),
                ("ecc_low_pl", N_pl),
                ("ecc_scale_g", N_g),
                ("ecc_scale_pl", N_pl),
            ]
        )

        error_magazine.register(
            Parameters.ECCENTRICITY.value,
            partial(
                truncated_normal_error,
                scale=err_params_value["ecc_scale"],
                low=err_params_value.get("ecc_low"),
                high=err_params_value.get("ecc_high"),
                cut_low=0.0,
                cut_high=1.0,
            ),
        )

    if has_mean_anomaly:
        parameters_name += (Parameters.MEAN_ANOMALY.value,)

        all_params.extend(
            [
                (Parameters.MEAN_ANOMALY.value + "_high_g", N_g),
                (Parameters.MEAN_ANOMALY.value + "_high_pl", N_pl),
                (Parameters.MEAN_ANOMALY.value + "_low_g", N_g),
                (Parameters.MEAN_ANOMALY.value + "_low_pl", N_pl),
            ]
        )

        error_magazine.register(
            Parameters.MEAN_ANOMALY.value,
            partial(
                truncated_normal_error,
                scale=err_params_value["mean_anomaly_scale"],
                low=err_params_value.get("mean_anomaly_low"),
                high=err_params_value.get("mean_anomaly_high"),
                cut_low=0.0,
                cut_high=1.0,
            ),
        )

    if has_redshift:
        parameters_name += (Parameters.REDSHIFT.value,)

        all_params.extend(
            [
                ("redshift_kappa_g", N_g),
                ("redshift_kappa_pl", N_pl),
                ("redshift_z_max_g", N_g),
                ("redshift_z_max_pl", N_pl),
            ]
        )

        error_magazine.register(
            Parameters.REDSHIFT.value,
            partial(
                truncated_normal_error,
                scale=err_params_value["redshift_scale"],
                low=err_params_value.get("redshift_low"),
                high=err_params_value.get("redshift_high"),
                cut_low=1e-3,
                cut_high=None,
            ),
        )

    if has_right_ascension:
        parameters_name += (Parameters.RIGHT_ASCENSION.value,)

        all_params.extend(
            [
                (Parameters.RIGHT_ASCENSION.value + "_high_g", N_g),
                (Parameters.RIGHT_ASCENSION.value + "_high_pl", N_pl),
                (Parameters.RIGHT_ASCENSION.value + "_low_g", N_g),
                (Parameters.RIGHT_ASCENSION.value + "_low_pl", N_pl),
            ]
        )

        error_magazine.register(
            Parameters.RIGHT_ASCENSION.value,
            partial(
                truncated_normal_error,
                scale=err_params_value[Parameters.RIGHT_ASCENSION.value + "_scale"],
                low=err_params_value.get(Parameters.RIGHT_ASCENSION.value + "_low"),
                high=err_params_value.get(Parameters.RIGHT_ASCENSION.value + "_high"),
                cut_low=0.0,
                cut_high=2.0 * jnp.pi,
            ),
        )

    if has_sin_declination:
        parameters_name += (Parameters.SIN_DECLINATION.value,)

        all_params.extend(
            [
                (Parameters.SIN_DECLINATION.value + "_high_g", N_g),
                (Parameters.SIN_DECLINATION.value + "_high_pl", N_pl),
                (Parameters.SIN_DECLINATION.value + "_low_g", N_g),
                (Parameters.SIN_DECLINATION.value + "_low_pl", N_pl),
            ]
        )

        error_magazine.register(
            Parameters.SIN_DECLINATION.value,
            partial(
                truncated_normal_error,
                scale=err_params_value[Parameters.SIN_DECLINATION.value + "_scale"],
                low=err_params_value.get(Parameters.SIN_DECLINATION.value + "_low"),
                high=err_params_value.get(Parameters.SIN_DECLINATION.value + "_high"),
                cut_low=-1.0,
                cut_high=1.0,
            ),
        )

    if has_detection_time:
        parameters_name += (Parameters.DETECTION_TIME.value,)

        all_params.extend(
            [
                (Parameters.DETECTION_TIME.value + "_high_g", N_g),
                (Parameters.DETECTION_TIME.value + "_high_pl", N_pl),
                (Parameters.DETECTION_TIME.value + "_low_g", N_g),
                (Parameters.DETECTION_TIME.value + "_low_pl", N_pl),
            ]
        )

        @error_magazine.register(Parameters.DETECTION_TIME.value)
        def detection_time_error(x, size, key):
            eps = 1e-6  # To avoid log(0) or log of negative
            safe_x = jnp.maximum(x, eps)
            err_x = dist.LogNormal(
                loc=jnp.log(safe_x),
                scale=err_params_value[Parameters.DETECTION_TIME.value + "_scale"],
            ).sample(key=key, sample_shape=(size,))

            return err_x

    if has_cos_iota:
        parameters_name += (Parameters.COS_IOTA.value,)

        all_params.extend(
            [
                (Parameters.COS_IOTA.value + "_high_g", N_g),
                (Parameters.COS_IOTA.value + "_high_pl", N_pl),
                (Parameters.COS_IOTA.value + "_low_g", N_g),
                (Parameters.COS_IOTA.value + "_low_pl", N_pl),
            ]
        )

        error_magazine.register(
            Parameters.COS_IOTA.value,
            partial(
                truncated_normal_error,
                scale=err_params_value[Parameters.COS_IOTA.value + "_scale"],
                low=err_params_value.get(Parameters.COS_IOTA.value + "_low"),
                high=err_params_value.get(Parameters.COS_IOTA.value + "_high"),
                cut_low=-1.0,
                cut_high=1.0,
            ),
        )

    if has_polarization_angle:
        parameters_name += (Parameters.POLARIZATION_ANGLE.value,)

        all_params.extend(
            [
                (Parameters.POLARIZATION_ANGLE.value + "_high_g", N_g),
                (Parameters.POLARIZATION_ANGLE.value + "_high_pl", N_pl),
                (Parameters.POLARIZATION_ANGLE.value + "_low_g", N_g),
                (Parameters.POLARIZATION_ANGLE.value + "_low_pl", N_pl),
            ]
        )

        error_magazine.register(
            Parameters.POLARIZATION_ANGLE.value,
            partial(
                truncated_normal_error,
                scale=err_params_value[Parameters.POLARIZATION_ANGLE.value + "_scale"],
                low=err_params_value.get(Parameters.POLARIZATION_ANGLE.value + "_low"),
                high=err_params_value.get(
                    Parameters.POLARIZATION_ANGLE.value + "_high"
                ),
                cut_low=0.0,
                cut_high=jnp.pi,
            ),
        )

    if has_phi_orb:
        parameters_name += (Parameters.PHI_ORB.value,)

        all_params.extend(
            [
                (Parameters.PHI_ORB.value + "_high_g", N_g),
                (Parameters.PHI_ORB.value + "_high_pl", N_pl),
                (Parameters.PHI_ORB.value + "_low_g", N_g),
                (Parameters.PHI_ORB.value + "_low_pl", N_pl),
            ]
        )

        @error_magazine.register(Parameters.PHI_ORB.value)
        def phi_orb_error(x, size, key):
            err_x = dist.TruncatedNormal(
                loc=x,
                scale=err_params_value[Parameters.PHI_ORB.value + "_scale"],
                low=err_params_value.get(Parameters.PHI_ORB.value + "_low", 0.0),
                high=err_params_value.get(
                    Parameters.PHI_ORB.value + "_high", 2 * jnp.pi
                ),
            ).sample(key=key, sample_shape=(size,))

            err_x = jnp.mod(err_x, 2 * jnp.pi)
            return err_x

    extended_params = []
    for params in all_params:
        extended_params.extend(expand_arguments(*params))

    model_param = match_all(extended_params, model_json)

    model_param.update(
        {
            "N_pl": N_pl,
            "N_g": N_g,
            "use_spin": has_spin,
            "use_tilt": has_tilt,
            "use_eccentricity": has_eccentricity,
            "use_mean_anomaly": has_mean_anomaly,
            "use_redshift": has_redshift,
            "use_cos_iota": has_cos_iota,
            "use_phi_12": has_phi_12,
            "use_polarization_angle": has_polarization_angle,
            "use_right_ascension": has_right_ascension,
            "use_sin_declination": has_sin_declination,
            "use_detection_time": has_detection_time,
            "use_phi_1": has_phi_1,
            "use_phi_2": has_phi_2,
            "use_phi_orb": has_phi_orb,
        }
    )

    nvt = vt_json_read_and_process(parameters_name, args.vt_json)
    log_selection_fn = nvt.get_mapped_logVT()

    pmean_key, factory_key = jrd.split(jrd.PRNGKey(args.seed), 2)

    pmean_kwargs = poisson_mean_parser.poisson_mean_parser(args.pmean_json)
    erate_estimator = PoissonMean(nvt, key=pmean_key, **pmean_kwargs)

    popfactory = PopulationFactory(
        model_fn=NPowerlawMGaussian,
        model_params=model_param,
        parameters=parameters_name,
        log_selection_fn=log_selection_fn,
        ERate_obj=erate_estimator,
        num_realizations=args.num_realizations,
        error_size=args.error_size,
    )
    popfactory.produce(factory_key)



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from collections.abc import Callable
from typing import Dict, List, Optional

import equinox as eqx
import jax
import numpyro
from jax import Array, numpy as jnp
from jaxtyping import ArrayLike
from numpyro._typing import DistributionT
from numpyro.distributions import Distribution

from ..models.utils import JointDistribution
from ..poisson_mean import PoissonMean


__all__ = ["numpyro_poisson_likelihood"]


@eqx.filter_jit
def _batch_log_prob(
    model_instance: DistributionT,
    batched_data: Array,
    batched_log_ref_priors: Array,
    batched_masks: Array,
) -> Array:
    safe_data = jnp.where(
        jnp.expand_dims(batched_masks, axis=-1),
        batched_data,
        model_instance.support.feasible_like(batched_data),
    )
    safe_log_ref_prior = jnp.where(batched_masks, batched_log_ref_priors, 0.0)

    batched_model_log_prob = jax.vmap(jax.vmap(model_instance.log_prob))(safe_data)  # type: ignore
    safe_model_log_prob = jnp.where(batched_masks, batched_model_log_prob, -jnp.inf)
    batched_log_prob: Array = safe_model_log_prob - safe_log_ref_prior
    batched_log_prob = jnp.where(
        batched_masks & (~jnp.isnan(batched_log_prob)),
        batched_log_prob,
        -jnp.inf,
    )
    log_prob_sum = jax.nn.logsumexp(
        batched_log_prob,
        axis=-1,
        where=~jnp.isneginf(batched_log_prob),
    )
    safe_log_prob_sum = jnp.where(jnp.isneginf(log_prob_sum), -jnp.inf, log_prob_sum)
    return jnp.sum(safe_log_prob_sum, axis=-1)


def numpyro_poisson_likelihood(
    dist_fn: Callable[..., DistributionT],
    priors: JointDistribution,
    variables: Dict[str, DistributionT],
    variables_index: Dict[str, int],
    log_constants: ArrayLike,
    ERate_obj: PoissonMean,
    where_fns: Optional[List[Callable[..., Array]]],
    constants: Dict[str, Array],
) -> Callable[[List[Array], List[Array], List[Array]], Array]:
    del priors

    def log_likelihood_fn(
        data_group: List[Array],
        log_ref_priors_group: List[Array],
        masks_group: List[Array],
    ):
        variables_samples = [
            numpyro.sample(parameter_name, prior_dist)
            for parameter_name, prior_dist in sorted(
                variables.items(), key=lambda x: x[0]
            )
        ]
        mapped_params = {
            name: variables_samples[i] for name, i in variables_index.items()
        }

        model_instance: Distribution = dist_fn(**mapped_params, validate_args=True)

        # μ = E_{θ|Λ}[VT(θ)]
        expected_rates = ERate_obj(model_instance)

        total_log_likelihood = log_constants  # - Σ log(M_i)
        # Σ log Σ exp (log p(θ|data_n) - log π_n)
        for batched_data, batched_log_ref_priors, batched_masks in zip(
            data_group, log_ref_priors_group, masks_group
        ):
            total_log_likelihood += _batch_log_prob(
                model_instance, batched_data, batched_log_ref_priors, batched_masks
            )

        log_likelihood = total_log_likelihood - expected_rates

        if where_fns is not None and len(where_fns) > 0:
            mask = where_fns[0](**constants, **mapped_params)
            for where_fn in where_fns[1:]:
                mask = mask & where_fn(**constants, **mapped_params)
            log_likelihood = jnp.where(
                mask,
                log_likelihood,
                -jnp.inf,  # type: ignore
            )

        # - μ + Σ log Σ exp (log p(θ|data_n) - log π_n) - Σ log(M_i)
        numpyro.factor("log_likelihood", log_likelihood)

    return log_likelihood_fn  # type: ignore



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from collections.abc import Sequence
from typing import Literal, Optional, Tuple

from jax import lax, numpy as jnp, random as jrd
from jaxtyping import Array, PRNGKeyArray
from numpyro.distributions import constraints, Distribution
from numpyro.distributions.util import validate_sample
from numpyro.util import is_prng_key

from ...utils.tools import error_if
from ..constraints import all_constraint


class JointDistribution(Distribution):
    pytree_aux_fields = ("shaped_values",)
    pytree_data_fields = ("_support", "marginal_distributions")

    def __init__(
        self,
        *marginal_distributions: Distribution,
        flatten_method: Optional[Literal["deep", "shallow"]] = None,
        support: Optional[constraints.Constraint] = None,
        validate_args: Optional[bool] = None,
    ) -> None:
        """Construct a joint distribution from one or more marginal distributions.

        You may pass individual `Distribution` instances or nest them inside
        :class:`JointDistribution`s. The `flatten_method` argument allows flattening of nested
        joints into a single flat list of marginals.

        Parameters
        ----------
        marginal_distributions : *Distribution
            One or more `Distribution` objects (or nested :class:`JointDistribution`s) that form
            the components of the joint distribution.
        flatten_method : Optional[Literal[&quot;deep&quot;, &quot;shallow&quot;]], optional
            If "shallow", one level of nested :class:`JointDistributions` will be flattened.
            If "deep", all levels of nested :class:`JointDistributions` will be recursively flattened.
            If None (default), the nesting is preserved as-is.
        support : Optional[constraints.Constraint], optional
            The constraint object representing the support of the joint distribution.
            If not provided, it is computed from the support of the marginals.
        validate_args : Optional[bool], optional
            Whether to validate distribution parameters and inputs. Default is None.

        Raises
        ------
        ValueError
             If no marginal distributions are provided.


        Example
        -------
        .. code::

            >>> from numpyro.distributions import Normal
            >>> from gwkokab.models.utils import JointDistribution

            >>> A = Normal(0, 1)
            >>> B = Normal(1, 1)
            >>> C = Normal(2, 1)
            >>> D = Normal(3, 1)
            >>> E = Normal(4, 1)

            >>> jd = JointDistribution(
            ...     A, JointDistribution(B, JointDistribution(C, D)), E
            ... )

            >>> len(jd.marginal_distributions)  # No flattening (default)
            3

            >>> jd = JointDistribution(
            ...     A,
            ...     JointDistribution(B, JointDistribution(C, D)),
            ...     E,
            ...     flatten_method="shallow",
            ... )
            >>> len(jd.marginal_distributions)  # Shallow flattening
            4

            >>> jd = JointDistribution(
            ...     A,
            ...     JointDistribution(B, JointDistribution(C, D)),
            ...     E,
            ...     flatten_method="deep",
            ... )
            >>> len(jd.marginal_distributions)  # Deep flattening
            5
        """
        error_if(
            not marginal_distributions,
            msg="At least one marginal distribution is required.",
        )

        marginal_flatten = _flatten_marginal_distributions(
            marginal_distributions, flatten_method
        )
        self.marginal_distributions: Sequence[Distribution] = tuple(marginal_flatten)
        self.shaped_values: Sequence[int | Tuple[int, int]] = tuple()
        batch_shape = lax.broadcast_shapes(
            *tuple(d.batch_shape for d in marginal_flatten)
        )
        k = 0
        for d in marginal_flatten:
            if d.event_shape:
                self.shaped_values += ((k, k + d.event_shape[0]),)
                k += d.event_shape[0]
            else:
                self.shaped_values += (k,)
                k += 1
        if support is None:
            self._support = all_constraint(
                [d.support for d in marginal_flatten], self.shaped_values
            )
        else:
            self._support = support
        super(JointDistribution, self).__init__(
            batch_shape=batch_shape,
            event_shape=(k,),
            validate_args=validate_args,
        )

    @constraints.dependent_property(is_discrete=False)
    def support(self) -> constraints.Constraint:
        """The support of the joint distribution."""
        return self._support

    @validate_sample
    def log_prob(self, value: Array) -> Array:
        log_prob_val = jnp.zeros(value.shape[:-1], dtype=value.dtype)
        for m_dist, event_slice in zip(self.marginal_distributions, self.shaped_values):
            if isinstance(event_slice, int):
                value_slice = lax.dynamic_index_in_dim(
                    value, event_slice, axis=-1, keepdims=False
                )
            else:
                value_slice = lax.dynamic_slice_in_dim(
                    value,
                    event_slice[0],
                    event_slice[1] - event_slice[0],
                    axis=-1,
                )
            log_prob_val += m_dist.log_prob(value_slice)
        return log_prob_val

    def sample(self, key: PRNGKeyArray, sample_shape: tuple[int, ...] = ()):
        assert is_prng_key(key)
        keys = tuple(jrd.split(key, len(self.marginal_distributions)))
        samples = [
            d.sample(k, sample_shape).reshape(*sample_shape, -1)
            for d, k in zip(self.marginal_distributions, keys)
        ]
        samples_concatenated = jnp.concatenate(samples, axis=-1)
        return samples_concatenated


def _flatten_marginal_distributions(
    marginal_distributions: Sequence[Distribution | JointDistribution],
    flatten_method: Optional[Literal["deep", "shallow"]] = None,
) -> Sequence[Distribution | JointDistribution]:
    """Flatten marginal distributions based on the specified method.

    This function converts nested :class:`JointDistribution` structures into a flat sequence
    of `Distribution` objects, depending on the chosen `flatten_method`

    Parameters
    ----------
    marginal_distributions : Sequence[Distribution  |  JointDistribution]
        Sequence of Distribution or JointDistribution instances.
    flatten_method : Optional[Literal[&quot;deep&quot;, &quot;shallow&quot;]], optional
        If "shallow", flattens one level of nesting.\n
        If "deep", recursively flattens all levels.\n
        If None, returns the input as-is., by default None

    Returns
    -------
    Sequence[Distribution | JointDistribution]
        A flattened sequence of distributions.

    Example
    -------
        Given:
            JointDistribution(A, JointDistribution(B, JointDistribution(C, D)), E)

        - flatten_method=None
            => (A, JointDistribution(B, JointDistribution(C, D)), E)

        - flatten_method='shallow'
            => (A, B, JointDistribution(C, D), E)

        - flatten_method='deep'
            => (A, B, C, D, E)
    """
    error_if(
        flatten_method not in (None, "deep", "shallow"),
        msg=f"Unknown flatten method: {flatten_method}",
    )
    if flatten_method is None:
        return marginal_distributions
    flatten_dists: list[Distribution] = []
    for m_dist in marginal_distributions:
        if isinstance(m_dist, JointDistribution):
            if flatten_method == "shallow":
                m_dist_marginal_distributions = m_dist.marginal_distributions
            else:  # deep case
                m_dist_marginal_distributions = _flatten_marginal_distributions(
                    m_dist.marginal_distributions, flatten_method
                )
            flatten_dists.extend(m_dist_marginal_distributions)
        else:
            flatten_dists.append(m_dist)
    return flatten_dists
